# This file is auto-generated with util/get_builtin_functions
# for MATLAB release R2017b

HIGH_IMPACT_BUILTIN_FUNCTIONS = frozenset([
    "Inf",
    "NaN",
    "cd",
    "clear",
    "clock",
    "close",
    "cputime",
    "double",
    "eps",
    "exit",
    "eye",
    "false",
    "flintmax",
    "horzcat",
    "inf",
    "int16",
    "int32",
    "int64",
    "int8",
    "isdeployed",
    "lasterr",
    "lastwarn",
    "license",
    "more",
    "nan",
    "nargin",
    "nargout",
    "newline",
    "ones",
    "pi",
    "quit",
    "rand",
    "randn",
    "realmax",
    "realmin",
    "single",
    "speye",
    "tic",
    "toc",
    "true",
    "uint16",
    "uint32",
    "uint64",
    "uint8",
    "vertcat",
    "zeros",
])

BUILTIN_FUNCTIONS = frozenset([
    "Inf",
    "NaN",
    "abs",
    "accumarray",
    "acos",
    "acosd",
    "acosh",
    "acot",
    "acotd",
    "acoth",
    "acsc",
    "acscd",
    "acsch",
    "actxGetRunningServer",
    "actxcontrol",
    "airy",
    "all",
    "amd",
    "ancestor",
    "and",
    "animatedline",
    "ans",
    "any",
    "arith",
    "arrayfun",
    "asec",
    "asecd",
    "asech",
    "asin",
    "asind",
    "asinh",
    "assert",
    "assignin",
    "atan",
    "atan2",
    "atan2d",
    "atand",
    "atanh",
    "axes",
    "balance",
    "bandwidth",
    "beep",
    "besselh",
    "besseli",
    "besselj",
    "besselk",
    "bessely",
    "betainc",
    "betaincinv",
    "bitand",
    "bitcmp",
    "bitget",
    "bitor",
    "bitset",
    "bitshift",
    "bitxor",
    "bsxfun",
    "builtin",
    "calllib",
    "cameratoolbar",
    "cast",
    "cat",
    "cd",
    "ceil",
    "cell",
    "cell2struct",
    "cellfun",
    "changeNotification",
    "changeNotificationAdvanced",
    "char",
    "chol",
    "cholupdate",
    "circshift",
    "class",
    "clc",
    "clear",
    "clock",
    "close",
    "colamd",
    "colon",
    "colstyle",
    "complex",
    "computer",
    "conj",
    "contourc",
    "conv2",
    "convhull",
    "copyfile",
    "copyobj",
    "copyoptionsfcn",
    "cos",
    "cosd",
    "cosh",
    "cot",
    "cotd",
    "coth",
    "cputime",
    "csc",
    "cscd",
    "csch",
    "ctranspose",
    "cummax",
    "cummin",
    "cumprod",
    "cumsum",
    "dataread",
    "dbclear",
    "dbcont",
    "dbdown",
    "dbquit",
    "dbstack",
    "dbstatus",
    "dbstep",
    "dbstop",
    "dbtype",
    "dbup",
    "ddeadv",
    "ddeexec",
    "ddeinit",
    "ddepoke",
    "ddereq",
    "ddeterm",
    "ddeunadv",
    "deblank",
    "delaunay",
    "delete",
    "det",
    "diag",
    "diary",
    "diff",
    "dir",
    "disp",
    "display",
    "dissect",
    "dmperm",
    "double",
    "dragrect",
    "drawnow",
    "echo",
    "eig",
    "enumeration",
    "eps",
    "eq",
    "erf",
    "erfc",
    "erfcinv",
    "erfcx",
    "erfinv",
    "error",
    "etree",
    "eval",
    "evalc",
    "evalin",
    "events",
    "exist",
    "exit",
    "exp",
    "expm1",
    "eye",
    "false",
    "fclose",
    "feof",
    "ferror",
    "feval",
    "fft",
    "fftn",
    "fftw",
    "fgets",
    "fieldnames",
    "figure",
    "fileattrib",
    "filemarker",
    "fill",
    "fill3",
    "filter",
    "find",
    "findall",
    "findobj",
    "findstr",
    "fix",
    "flintmax",
    "flip",
    "floor",
    "fopen",
    "format",
    "fprintf",
    "frame2im",
    "fread",
    "fscanf",
    "fseek",
    "ftell",
    "full",
    "func2str",
    "functions",
    "fwrite",
    "gamma",
    "gammainc",
    "gammaincinv",
    "gammaln",
    "gcbo",
    "gcf",
    "gco",
    "ge",
    "get",
    "getappdata",
    "getenv",
    "getmcruserdata",
    "gobjects",
    "graphicsversion",
    "groot",
    "gt",
    "handle",
    "handle2struct",
    "hess",
    "hgclose",
    "hgconvertunits",
    "hgexport",
    "hggroup",
    "hgload",
    "hgtransform",
    "histc",
    "home",
    "horzcat",
    "hypot",
    "i",
    "ichol",
    "ifft",
    "ifftn",
    "ilu",
    "im2frame",
    "imag",
    "image",
    "inf",
    "inferiorto",
    "inmem",
    "input",
    "inputname",
    "int16",
    "int32",
    "int64",
    "int8",
    "inv",
    "isa",
    "isappdata",
    "isbanded",
    "iscell",
    "iscellstr",
    "ischar",
    "iscolumn",
    "isdeployed",
    "isdiag",
    "isempty",
    "isequal",
    "isequaln",
    "isequalwithequalnans",
    "isfield",
    "isfile",
    "isfinite",
    "isfloat",
    "isfolder",
    "isgraphics",
    "ishandle",
    "ishermitian",
    "ishghandle",
    "isinf",
    "isinteger",
    "isjava",
    "isletter",
    "islogical",
    "ismatrix",
    "ismcc",
    "ismembertol",
    "isnan",
    "isnumeric",
    "isplotchild",
    "isreal",
    "isrow",
    "isscalar",
    "issorted",
    "issortedrows",
    "isspace",
    "issparse",
    "isstr",
    "isstring",
    "isstrprop",
    "isstruct",
    "isstudent",
    "issymmetric",
    "istril",
    "istriu",
    "isvector",
    "j",
    "java",
    "javaArray",
    "javaMethod",
    "javaMethodEDT",
    "javaObject",
    "javaObjectEDT",
    "jpropeditutils",
    "jsondecode",
    "jsonencode",
    "keyboard",
    "lasterr",
    "lasterror",
    "lastwarn",
    "ldivide",
    "ldl",
    "le",
    "length",
    "libisloaded",
    "license",
    "light",
    "line",
    "linsolve",
    "lists",
    "load",
    "loadlibrary",
    "loadobj",
    "localfunctions",
    "log",
    "log10",
    "log1p",
    "log2",
    "logical",
    "loglog",
    "logo",
    "lookfor",
    "lower",
    "lt",
    "ltitr",
    "lu",
    "matlabpath",
    "matlabroot",
    "max",
    "maxk",
    "memory",
    "message",
    "metaclass",
    "methods",
    "mex",
    "mex_helper",
    "mexext",
    "mfilename",
    "min",
    "mink",
    "minus",
    "mislocked",
    "mkdir",
    "mldivide",
    "mlock",
    "mod",
    "more",
    "movefile",
    "movie",
    "movmad",
    "movmax",
    "movmean",
    "movmedian",
    "movmin",
    "movprod",
    "movstd",
    "movsum",
    "movvar",
    "mpower",
    "mrdivide",
    "mtimes",
    "munlock",
    "mwsamp",
    "namelengthmax",
    "nan",
    "nargchk",
    "nargin",
    "narginchk",
    "nargout",
    "nargoutchk",
    "native2unicode",
    "ndims",
    "ne",
    "newline",
    "newplot",
    "nnz",
    "nonzeros",
    "norm",
    "not",
    "numArgumentsFromSubscript",
    "numel",
    "nzmax",
    "ones",
    "or",
    "ordeig",
    "ordqz",
    "ordschur",
    "pack",
    "paren",
    "partialpath",
    "patch",
    "pause",
    "permute",
    "pi",
    "plot",
    "plot3",
    "plotyy",
    "plus",
    "polaraxes",
    "pow2",
    "power",
    "precedence",
    "prod",
    "properties",
    "psi",
    "punct",
    "py",
    "pyargs",
    "pyversion",
    "qr",
    "qrupdate",
    "quit",
    "qz",
    "rand",
    "randi",
    "randn",
    "randperm",
    "rbbox",
    "rcond",
    "rdivide",
    "real",
    "reallog",
    "realmax",
    "realmin",
    "realpow",
    "realsqrt",
    "rectangle",
    "recycle",
    "regexp",
    "regexpi",
    "regexprep",
    "regexptranslate",
    "rehash",
    "relop",
    "rem",
    "repelem",
    "repmat",
    "reset",
    "reshape",
    "rethrow",
    "rmappdata",
    "rmdir",
    "round",
    "save",
    "savefig",
    "saveobj",
    "schur",
    "script",
    "sec",
    "secd",
    "sech",
    "selectmoveresize",
    "semilogx",
    "semilogy",
    "seriallist",
    "set",
    "setappdata",
    "setenv",
    "setmcruserdata",
    "setstr",
    "setupplotbrowser",
    "sign",
    "sin",
    "sind",
    "single",
    "sinh",
    "size",
    "slash",
    "sort",
    "sparse",
    "speye",
    "sprintf",
    "sqrt",
    "sscanf",
    "str2func",
    "strcmp",
    "strcmpi",
    "strfind",
    "strings",
    "strncmp",
    "strncmpi",
    "strrep",
    "strtrim",
    "struct",
    "struct2cell",
    "struct2handle",
    "structfun",
    "subsasgn",
    "subsindex",
    "subsref",
    "sum",
    "superclasses",
    "superiorto",
    "support",
    "surface",
    "svd",
    "symamd",
    "symbfact",
    "symrcm",
    "syntax",
    "system",
    "tan",
    "tand",
    "tanh",
    "tex",
    "text",
    "textscan",
    "tic",
    "times",
    "toc",
    "transpose",
    "tril",
    "triu",
    "true",
    "type",
    "typecast",
    "uicontainer",
    "uicontextmenu",
    "uicontrol",
    "uiflowcontainer",
    "uigetdir",
    "uigetfile",
    "uigridcontainer",
    "uimenu",
    "uint16",
    "uint32",
    "uint64",
    "uint8",
    "uipanel",
    "uipushtool",
    "uiputfile",
    "uisetcolor",
    "uisetfont",
    "uitable",
    "uitoggletool",
    "uitoolbar",
    "uminus",
    "unicode2native",
    "uniquetol",
    "unloadlibrary",
    "uplus",
    "upper",
    "varargin",
    "varargout",
    "vecnorm",
    "vertcat",
    "vissuite",
    "waitfor",
    "waitforbuttonpress",
    "warning",
    "what",
    "which",
    "who",
    "whos",
    "winqueryreg",
    "xor",
    "yyaxis",
    "zeros",
    "zoom",
])

EXTRA_FUNCTIONS = frozenset([
    "NaT",
    "activateuimode",
    "actxcontrolcreateproperty",
    "actxcontrollist",
    "actxcontrolselect",
    "actxproxy",
    "actxserver",
    "adddatadlg",
    "addpath",
    "addpref",
    "addsubplot",
    "addtodate",
    "align",
    "alim",
    "allMuPADNotebooks",
    "allchild",
    "alpha",
    "alphamap",
    "analyzeCodeCompatibility",
    "angle",
    "annotation",
    "appdesigner",
    "applicationCompiler",
    "arduino",
    "area",
    "array2table",
    "array2timetable",
    "arrayviewfunc",
    "audiodevinfo",
    "audioinfo",
    "audioread",
    "audiouniquename",
    "audiowrite",
    "auditcontents",
    "automesh",
    "autumn",
    "avgate",
    "avifinfo",
    "aviinfo",
    "awtcreate",
    "awtinvoke",
    "axescheck",
    "axis",
    "bar",
    "bar3",
    "bar3h",
    "barh",
    "base2dec",
    "basicfitdatastat",
    "bench",
    "beta",
    "betaln",
    "bicg",
    "bicgstab",
    "bicgstabl",
    "bin2dec",
    "binscatter",
    "blanks",
    "blkdiag",
    "bone",
    "boundary",
    "bounds",
    "box",
    "brighten",
    "brush",
    "btndown",
    "btngroup",
    "btnicon",
    "btnpress",
    "btnresize",
    "btnstate",
    "btnup",
    "builddocsearchdb",
    "bvp4c",
    "bvp5c",
    "bvpget",
    "bvpinit",
    "bvpset",
    "bvpxtend",
    "bwcontr",
    "caldays",
    "calendar",
    "callAllOptimOutputFcns",
    "callAllOptimPlotFcns",
    "callSoapService",
    "callgraphviz",
    "calmonths",
    "calquarters",
    "calweeks",
    "calyears",
    "camdolly",
    "cameraCalibrator",
    "cameramenu",
    "cameratoolbar",
    "camlight",
    "camlookat",
    "camorbit",
    "campan",
    "campos",
    "camproj",
    "camroll",
    "camrotate",
    "camtarget",
    "camup",
    "camva",
    "camzoom",
    "capturescreen",
    "cart2pol",
    "cart2sph",
    "caxis",
    "cdf2rdf",
    "cdfinfo",
    "cdfread",
    "cdfwrite",
    "cell2mat",
    "cell2table",
    "celldisp",
    "cellplot",
    "cellstr",
    "cftool",
    "cgs",
    "changeseriestype",
    "checkSyntacticWarnings",
    "checkcode",
    "checkin",
    "checkout",
    "cla",
    "clabel",
    "classificationLearner",
    "clearAllMemoizedCaches",
    "clearvars",
    "clf",
    "clipboard",
    "closereq",
    "cmapeditor",
    "cmopts",
    "cmpermute",
    "cmunique",
    "code2html",
    "codeCompatibilityReport",
    "codegen",
    "codetoolsswitchyard",
    "colorThresholder",
    "colorbar",
    "colorcube",
    "colordef",
    "colormap",
    "colormapeditor",
    "colornone",
    "colorpalette",
    "colperm",
    "comet",
    "comet3",
    "commandhistory",
    "commandwindow",
    "commonplotfunc",
    "compan",
    "compass",
    "compose",
    "cond",
    "condeig",
    "condest",
    "coneplot",
    "connector",
    "consume_assign",
    "contains",
    "contentsrpt",
    "contour",
    "contour3",
    "contourf",
    "contours",
    "contourslice",
    "contrast",
    "controlSystemTuner",
    "conv",
    "convertCharsToStrings",
    "convertSpreadsheetDates",
    "convertSpreadsheetExcelDates",
    "convertStringsToChars",
    "convhulln",
    "convn",
    "cool",
    "copper",
    "corrcoef",
    "count",
    "cov",
    "coveragerpt",
    "cplxpair",
    "createCellArrayOfFunctions",
    "createClassFromWsdl",
    "createProblemStruct",
    "createSoapMessage",
    "createSolverStruct",
    "cross",
    "cshelp",
    "csvread",
    "csvwrite",
    "cumtrapz",
    "curl",
    "customverctrl",
    "cylinder",
    "daexplr",
    "daqread",
    "daspect",
    "datachildren",
    "datacursormode",
    "datastore",
    "datatipinfo",
    "date",
    "dateTickPicker",
    "datenum",
    "datestr",
    "datetick",
    "datetickstr",
    "datevec",
    "days",
    "dblquad",
    "dbmex",
    "dde23",
    "ddeget",
    "ddensd",
    "ddesd",
    "ddeset",
    "deal",
    "dec2base",
    "dec2bin",
    "dec2hex",
    "decic",
    "deconv",
    "defaultItemHitCallback",
    "defaulterrorcallback",
    "deg2rad",
    "del2",
    "delaunayn",
    "delsq",
    "demo",
    "demos",
    "demowin",
    "depfunprophelper",
    "deployprefdir",
    "deploytool",
    "deprpt",
    "deselectall",
    "desktop",
    "desktopmenufcn",
    "details",
    "detrend",
    "deval",
    "dialog",
    "diary",
    "diffuse",
    "discretize",
    "displayEndOfDemoMessage",
    "dither",
    "divergence",
    "dlmread",
    "dlmwrite",
    "doc",
    "doclick",
    "doclink",
    "docpath",
    "docroot",
    "docsearch",
    "dofixrpt",
    "dokeypress",
    "domymenu",
    "doresize",
    "dos",
    "dot",
    "dsearchn",
    "dspfwiz",
    "echodemo",
    "edit",
    "editmenufcn",
    "eigs",
    "ellipj",
    "ellipke",
    "ellipsoid",
    "enableNETfromNetworkDrive",
    "enableplottoolbuttons",
    "enableservice",
    "enddrag",
    "endsWith",
    "eomday",
    "erase",
    "eraseBetween",
    "errorbar",
    "errordlg",
    "etime",
    "etreeplot",
    "eventlisteners",
    "exit",
    "expint",
    "expm",
    "export2wsdlg",
    "exportsetupdlg",
    "extractAfter",
    "extractBefore",
    "extractBetween",
    "ezcontour",
    "ezcontourf",
    "ezgraph3",
    "ezmesh",
    "ezmeshc",
    "ezplot",
    "ezplot3",
    "ezpolar",
    "ezsurf",
    "ezsurfc",
    "factor",
    "factorial",
    "fcnchk",
    "fcontour",
    "feather",
    "fft2",
    "fftshift",
    "fgetl",
    "figcopytemplatelistener",
    "figcopytemplatelistener_legacy",
    "figflag",
    "fignamer",
    "figtoolset",
    "figureToolbarCreateFcn",
    "figuredeletedlistener",
    "figureheaderdlg",
    "figurepalette",
    "fileDatastore",
    "filebrowser",
    "filemenufcn",
    "fileparts",
    "fileread",
    "filesep",
    "fillmissing",
    "filloutliers",
    "filter2",
    "filterBuilder",
    "fimplicit",
    "fimplicit3",
    "findExample",
    "findall",
    "findfigs",
    "findgroups",
    "findprinters",
    "finfo",
    "fitsdisp",
    "fitsinfo",
    "fitsread",
    "fitswrite",
    "fixcontents",
    "fixquote",
    "flag",
    "flipdim",
    "fliplr",
    "flipud",
    "fmesh",
    "fminbnd",
    "fminsearch",
    "fplot",
    "fplot3",
    "freqspace",
    "frewind",
    "fsurf",
    "fullfile",
    "function_handle",
    "functionhintsfunc",
    "functiontests",
    "funm",
    "fzero",
    "gallery",
    "gather",
    "gca",
    "gcbf",
    "gcd",
    "gcmr",
    "generateProfilerSelect",
    "genpath",
    "genvarname",
    "geobubble",
    "geolimits",
    "getFigureToolManager",
    "getcallinfo",
    "getcolumn",
    "getcrosssection",
    "getfield",
    "getfigurefordesktopclient",
    "getframe",
    "getobj",
    "getorcreateobj",
    "getpixelposition",
    "getplotbrowserproptable",
    "getplottool",
    "getpref",
    "getprinttemplate",
    "getptr",
    "getrangefromclass",
    "getscribecontextmenu",
    "getscribeobjectdata",
    "getselectobjects",
    "getstatus",
    "getuimode",
    "ginput",
    "gmres",
    "gobjects",
    "gplot",
    "grabcode",
    "gradient",
    "graph2dhelper",
    "gray",
    "graymon",
    "grid",
    "griddata",
    "griddatan",
    "gsvd",
    "gtext",
    "gui_mainfcn",
    "guidata",
    "guide",
    "guidefunc",
    "guidemfile",
    "guideopts",
    "guidetemplate",
    "guidetemplate0",
    "guidetemplate1",
    "guidetemplate2",
    "guidetemplate3",
    "guidetoolfunc",
    "guihandles",
    "gunzip",
    "gzip",
    "h5create",
    "h5disp",
    "h5info",
    "h5read",
    "h5readatt",
    "h5write",
    "h5writeatt",
    "hadamard",
    "hankel",
    "hardcopy",
    "hasbehavior",
    "hasuimode",
    "hdf",
    "hdf5",
    "hdf5info",
    "hdf5read",
    "hdf5write",
    "hdfan",
    "hdfdf24",
    "hdfdfr8",
    "hdfgd",
    "hdfh",
    "hdfhd",
    "hdfhe",
    "hdfhx",
    "hdfinfo",
    "hdfml",
    "hdfpt",
    "hdfread",
    "hdfsd",
    "hdfsw",
    "hdftool",
    "hdfv",
    "hdfvf",
    "hdfvh",
    "hdfvs",
    "heatmap",
    "help",
    "help2html",
    "helpPopup",
    "helpbrowser",
    "helpdesk",
    "helpdlg",
    "helpmenufcn",
    "helprpt",
    "helpview",
    "helpwin",
    "hex2dec",
    "hex2num",
    "hgaddbehavior",
    "hgbehaviorfactory",
    "hgdisp",
    "hgfeval",
    "hggetbehavior",
    "hggetdisp",
    "hgline2lineseries",
    "hgload",
    "hgsave",
    "hgsetdisp",
    "hidden",
    "hilb",
    "hist",
    "histcounts",
    "histcounts2",
    "histogram",
    "histogram2",
    "hold",
    "home",
    "hot",
    "hours",
    "hsv",
    "hsv2rgb",
    "icondisp",
    "iconeditor",
    "iconread",
    "idivide",
    "ifft2",
    "ifftshift",
    "im2double",
    "im2java",
    "imageBatchProcessor",
    "imageDatastore",
    "imageRegionAnalyzer",
    "imageSegmenter",
    "imagesc",
    "imageview",
    "imapprox",
    "imfinfo",
    "imformats",
    "implay",
    "importdata",
    "imread",
    "imresize",
    "imshow",
    "imwrite",
    "ind2rgb",
    "ind2sub",
    "indentcode",
    "indexhelper",
    "info",
    "initdesktoputils",
    "inlineeval",
    "inpolygon",
    "input",
    "inputdlg",
    "insertAfter",
    "insertBefore",
    "insertmenufcn",
    "inspect",
    "instrcb",
    "instrfind",
    "instrfindall",
    "int2str",
    "integral",
    "integral2",
    "integral3",
    "interp1",
    "interp1q",
    "interp2",
    "interp3",
    "interpft",
    "interpn",
    "interpstreamspeed",
    "intersect",
    "intmax",
    "intmin",
    "invhilb",
    "invokeProfiler",
    "ipcam",
    "ipermute",
    "is2D",
    "isStringScalar",
    "isactiveuimode",
    "iscalendarduration",
    "iscategorical",
    "ischange",
    "iscom",
    "isdatetime",
    "isdicom",
    "isdir",
    "isdpx",
    "isduration",
    "isenum",
    "isevent",
    "ishold",
    "isinterface",
    "iskeyword",
    "islocalmax",
    "islocalmin",
    "ismac",
    "ismember",
    "ismethod",
    "ismissing",
    "isnitf",
    "isobject",
    "isocaps",
    "isocolors",
    "isonormals",
    "isosurface",
    "isoutlier",
    "ispc",
    "isplotchild",
    "ispref",
    "isprime",
    "isprop",
    "istable",
    "istall",
    "istimetable",
    "isunix",
    "isvarname",
    "javaAddLsnrsToFigure",
    "javaGetHandles",
    "javaaddpath",
    "javachk",
    "javaclasspath",
    "javacomponent",
    "javarmpath",
    "jet",
    "join",
    "keyboard",
    "kron",
    "lcm",
    "legend",
    "legendcolorbarlayout",
    "legendinfo",
    "legendre",
    "libfunctions",
    "libfunctionsview",
    "libpointer",
    "libstruct",
    "lightangle",
    "lighting",
    "lin2mu",
    "lines",
    "linkaxes",
    "linkdata",
    "linkplotfunc",
    "linkprop",
    "linspace",
    "listdlg",
    "listfonts",
    "loadlibrary",
    "logm",
    "logspace",
    "ls",
    "lscov",
    "lsqminnorm",
    "lsqnonneg",
    "lsqr",
    "m2struct",
    "magic",
    "makePixelPath",
    "makebars",
    "makecontentsfile",
    "makedisplaynames",
    "makehgtform",
    "makemcode",
    "makemenu",
    "makeprofilerfooter",
    "makeprofilerheader",
    "mapreduce",
    "mapreducer",
    "mat2cell",
    "mat2str",
    "material",
    "matfile",
    "matfinfo",
    "maxNumCompThreads",
    "mdbfileonpath",
    "mdbpublish",
    "mdbstatus",
    "mean",
    "median",
    "membrane",
    "memoize",
    "menu",
    "menueditfunc",
    "menulabel",
    "mesh",
    "meshc",
    "meshgrid",
    "meshz",
    "methodsview",
    "mex",
    "mex_legacy",
    "mexdebug",
    "middrag",
    "milliseconds",
    "minres",
    "minutes",
    "mkpp",
    "mlappfinfo",
    "mlappinstallfinfo",
    "mlint",
    "mlintrpt",
    "mltbxfinfo",
    "mltimerpackage",
    "mlxdoc",
    "mmfileinfo",
    "mobiledev",
    "mode",
    "more",
    "moveaxis",
    "movegui",
    "movie2avi",
    "moviein",
    "movieview",
    "msgbox",
    "mu2lin",
    "multibandread",
    "multibandwrite",
    "mupadwelcome",
    "mustBeFinite",
    "mustBeGreaterThan",
    "mustBeGreaterThanOrEqual",
    "mustBeInteger",
    "mustBeLessThan",
    "mustBeLessThanOrEqual",
    "mustBeMember",
    "mustBeNegative",
    "mustBeNonNan",
    "mustBeNonempty",
    "mustBeNonnegative",
    "mustBeNonpositive",
    "mustBeNonsparse",
    "mustBeNonzero",
    "mustBeNumeric",
    "mustBeNumericOrLogical",
    "mustBePositive",
    "mustBeReal",
    "nccreate",
    "ncdisp",
    "nchoosek",
    "ncinfo",
    "ncread",
    "ncreadatt",
    "ncwrite",
    "ncwriteatt",
    "ncwriteschema",
    "ndgrid",
    "nested",
    "nextpow2",
    "noanimate",
    "nodither",
    "normest",
    "normest1",
    "now",
    "nsidedpoly",
    "nthroot",
    "null",
    "num2cell",
    "num2hex",
    "num2ruler",
    "num2str",
    "numgrid",
    "numjac",
    "objbounds",
    "ocrTrainer",
    "ode113",
    "ode15i",
    "ode15s",
    "ode23",
    "ode23s",
    "ode23t",
    "ode23tb",
    "ode45",
    "odeexamples",
    "odeget",
    "odephas2",
    "odephas3",
    "odeplot",
    "odeprint",
    "odeset",
    "odextend",
    "online_concatenator",
    "open",
    "openExample",
    "openFigure",
    "open_system",
    "openfig",
    "opengl",
    "openmlapp",
    "openmn",
    "openmu",
    "opentoline",
    "openvar",
    "openxvc",
    "openxvz",
    "optimget",
    "optimplotfunccount",
    "optimplotfval",
    "optimplotx",
    "optimset",
    "orderfields",
    "orient",
    "orth",
    "overobj",
    "pad",
    "pagesetupdlg",
    "pan",
    "parallel_function",
    "pareto",
    "parfor_M_check",
    "parfor_endpoint_check",
    "parfor_range_check",
    "parfor_sliced_fcnhdl_check",
    "parseSoapResponse",
    "parseparams",
    "parseplotapi",
    "parula",
    "pascal",
    "path",
    "path2rc",
    "pathsep",
    "pathtool",
    "pause",
    "pbaspect",
    "pcg",
    "pchip",
    "pcode",
    "pcolor",
    "pdepe",
    "pdeval",
    "peaks",
    "perl",
    "perms",
    "pie",
    "pie3",
    "pink",
    "pinv",
    "planerot",
    "playshow",
    "plotbrowser",
    "plotchild",
    "plotdoneevent",
    "plotedit",
    "plotmatrix",
    "plotpickerfunc",
    "plottoolfunc",
    "plottools",
    "plotutils",
    "plotyy",
    "pol2cart",
    "polar",
    "polarhistogram",
    "polarplot",
    "polarscatter",
    "poly",
    "polyarea",
    "polyder",
    "polyeig",
    "polyfit",
    "polyint",
    "polyspaceBugFinder",
    "polyspaceCodeProver",
    "polyval",
    "polyvalm",
    "popupstr",
    "ppval",
    "prefdir",
    "preferences",
    "prefspanel",
    "prepdrag",
    "primes",
    "print",
    "printProfilerRefreshButton",
    "printdlg",
    "printfProfilerLink",
    "printjob",
    "printpreview",
    "printtables",
    "printtemplate",
    "prism",
    "profile",
    "profreport",
    "profsave",
    "profview",
    "profviewgateway",
    "projdumpmat",
    "propedit",
    "propertyeditor",
    "publish",
    "putdowntext",
    "pwch",
    "pwd",
    "qhull",
    "qmr",
    "qrdelete",
    "qrinsert",
    "quad",
    "quad2d",
    "quadgk",
    "quadl",
    "quadv",
    "questdlg",
    "quit",
    "quiver",
    "quiver3",
    "rad2deg",
    "rank",
    "raspberrypi",
    "raspi",
    "rat",
    "rats",
    "readMobileSensorData",
    "readtable",
    "rectint",
    "reducepatch",
    "reducevolume",
    "refresh",
    "refreshdata",
    "regionpicker",
    "registerevent",
    "regmatlabserver",
    "remapfig",
    "renameStructField",
    "replace",
    "replaceBetween",
    "rescale",
    "resetplotview",
    "reverse",
    "reverse_binary_operator",
    "rfBudgetAnalyzer",
    "rgb2gray",
    "rgb2hsv",
    "rgb2ind",
    "rgbplot",
    "ribbon",
    "rjr",
    "rlim",
    "rmfield",
    "rmmissing",
    "rmpath",
    "rmpref",
    "rng",
    "roots",
    "rose",
    "rosser",
    "rot90",
    "rotate",
    "rotate3d",
    "rref",
    "rsf2csf",
    "rtickangle",
    "rtickformat",
    "rticklabels",
    "rticks",
    "ruler2num",
    "run",
    "runperf",
    "runreport",
    "runtests",
    "sampev",
    "saveFigure",
    "saveas",
    "savefig",
    "savepath",
    "savtoner",
    "scatter",
    "scatter3",
    "scribealign",
    "scribeccp",
    "scribeclearmode",
    "scribecursors",
    "scribeeventhandler",
    "scriberestoresavefcns",
    "scribetextdlg",
    "seconds",
    "selectobject",
    "sendmail",
    "separateOptimStruct",
    "setProfilerHtmlText",
    "setdiff",
    "setfield",
    "setpixelposition",
    "setpref",
    "setprinttemplate",
    "setptr",
    "setscribecontextmenu",
    "setscribeobjectdata",
    "setstatus",
    "settings",
    "setupExample",
    "setxor",
    "shading",
    "sharedplotfunc",
    "shg",
    "shiftdim",
    "showAddonExplorer",
    "showAddonExplorerForHardwareSupportPackages",
    "showHelpPreferences",
    "showdemo",
    "showplottool",
    "shrinkfaces",
    "signalAnalyzer",
    "simbiology",
    "simulink",
    "simulinkts2ts",
    "slLibraryBrowser",
    "slice",
    "smooth3",
    "smoothdata",
    "snapnow",
    "snaptogrid",
    "snlsFixedVarEvalOutputOrPlotFcn",
    "snlsFixedVarWrapOutputAndPlotFcns",
    "sortrows",
    "sound",
    "soundsc",
    "soundview",
    "spalloc",
    "spaugment",
    "spconvert",
    "spdiags",
    "specgraphhelper",
    "specular",
    "spfun",
    "sph2cart",
    "sphere",
    "spinmap",
    "spline",
    "split",
    "splitapply",
    "splitlines",
    "splncore",
    "spmd_feval",
    "spones",
    "spparms",
    "sprand",
    "sprandn",
    "sprandsym",
    "sprank",
    "spreadsheetDatastore",
    "spring",
    "spy",
    "sqrtm",
    "squeeze",
    "stairs",
    "standardizeMissing",
    "startsWith",
    "startscribeobject",
    "startscribepinning",
    "stateflow",
    "std",
    "stem",
    "stem3",
    "stereoCameraCalibrator",
    "str2double",
    "str2mat",
    "str2num",
    "strcat",
    "stream2",
    "stream3",
    "streamline",
    "streamparticles",
    "streamribbon",
    "streamslice",
    "streamtube",
    "strip",
    "stripanchors",
    "strjoin",
    "strjust",
    "strlength",
    "strmatch",
    "strread",
    "strsplit",
    "strtok",
    "struct2table",
    "strvcat",
    "sub2ind",
    "subplot",
    "subspace",
    "substruct",
    "subvolume",
    "summer",
    "superiorfloat",
    "support",
    "supportPackageInstaller",
    "surf",
    "surf2patch",
    "surfc",
    "surfl",
    "surfnorm",
    "svds",
    "swapbytes",
    "sylvester",
    "symmlq",
    "symvar",
    "tabdlg",
    "table2array",
    "table2cell",
    "table2struct",
    "table2timetable",
    "tabularTextDatastore",
    "tallrng",
    "tar",
    "tempdir",
    "tempname",
    "testsuite",
    "tetramesh",
    "texlabel",
    "textread",
    "textwrap",
    "tfqmr",
    "thetalim",
    "thetatickformat",
    "thetaticklabels",
    "thetaticks",
    "thingSpeakArea",
    "thingSpeakAuthenticate",
    "thingSpeakAuthenticatedList",
    "thingSpeakClearAuthentication",
    "thingSpeakPlot",
    "thingSpeakPlotYY",
    "thingSpeakRead",
    "thingSpeakScatter",
    "thingSpeakStem",
    "thingSpeakWrite",
    "timeit",
    "timercb",
    "timerfind",
    "timerfindall",
    "timetable2table",
    "timezones",
    "tipoftheday",
    "title",
    "toeplitz",
    "toolbareditor",
    "toolboxdir",
    "toolpalette",
    "toolsmenufcn",
    "topkrows",
    "trace",
    "trainingImageLabeler",
    "trapz",
    "treelayout",
    "treeplot",
    "trimesh",
    "triplequad",
    "triplot",
    "trisurf",
    "tsAlignSizes",
    "tsAnalyzeAbsTime",
    "tsChkTime",
    "tsIsDateFormat",
    "tsIsSameTime",
    "tsarrayFcn",
    "tsdateinterval",
    "tsearchn",
    "tsfcncallrouter",
    "tsgetDateFormat",
    "tsgetrelativetime",
    "tsinterp",
    "tsnaniqr",
    "tsnanmax",
    "tsnanmean",
    "tsnanmedian",
    "tsnanmin",
    "tsnanmode",
    "tsnanstd",
    "tsnansum",
    "tsnanvar",
    "tsnewevent",
    "tspnmatch",
    "tsprctile",
    "tspvformat",
    "tssorttime",
    "tsunitconv",
    "uialert",
    "uiaxes",
    "uibutton",
    "uibuttongroup",
    "uicheckbox",
    "uiclearmode",
    "uiconfirm",
    "uicontrol",
    "uidropdown",
    "uieditfield",
    "uifigure",
    "uigauge",
    "uigetdir",
    "uigetfile",
    "uigetmodemanager",
    "uigetpref",
    "uigettool",
    "uigettoolbar",
    "uiimport",
    "uiknob",
    "uilabel",
    "uilamp",
    "uilistbox",
    "uiload",
    "uimode",
    "uiopen",
    "uiputfile",
    "uiradiobutton",
    "uirestore",
    "uiresume",
    "uisave",
    "uisetcolor",
    "uisetfont",
    "uisetpref",
    "uislider",
    "uispinner",
    "uistack",
    "uisuspend",
    "uiswitch",
    "uitab",
    "uitabgroup",
    "uitextarea",
    "uitogglebutton",
    "uitoolfactory",
    "uitree",
    "uitreenode",
    "uiundo",
    "uiwait",
    "undocheckout",
    "union",
    "unique",
    "unix",
    "unmesh",
    "unmkpp",
    "unregisterallevents",
    "unregisterevent",
    "untar",
    "unwrap",
    "unzip",
    "updateHelpPopup",
    "urldecode",
    "urlencode",
    "urlfilter",
    "urlread",
    "urlwrite",
    "usejava",
    "usejavacomponent",
    "usev6plotapi",
    "utTrimNans",
    "uttsget",
    "validOptimProblemStruct",
    "validateattributes",
    "validateopts_UseParallel",
    "validatestring",
    "value2CustomFcn",
    "vander",
    "var",
    "vectorize",
    "ver",
    "verLessThan",
    "verctrl",
    "vertexpicker",
    "vga",
    "view",
    "viewmenufcn",
    "viewmtx",
    "volumebounds",
    "voronoi",
    "voronoin",
    "waitbar",
    "warndlg",
    "waterfall",
    "web",
    "webcam",
    "webcamlist",
    "webmap",
    "webread",
    "websave",
    "webwrite",
    "weekday",
    "whatsnew",
    "white",
    "whitebg",
    "wilkinson",
    "winmenu",
    "winopen",
    "winter",
    "wizard",
    "wordcloud",
    "workspace",
    "workspacefunc",
    "writeactxlicense",
    "writetable",
    "xlabel",
    "xlim",
    "xlsfinfo",
    "xlsread",
    "xlswrite",
    "xmlread",
    "xmlwrite",
    "xslt",
    "xtickangle",
    "xtickformat",
    "xticklabels",
    "xticks",
    "xychk",
    "xyzchk",
    "xyzvchk",
    "years",
    "ylabel",
    "ylim",
    "ytickangle",
    "ytickformat",
    "yticklabels",
    "yticks",
    "zip",
    "zlabel",
    "zlim",
    "zoom",
    "ztickangle",
    "ztickformat",
    "zticklabels",
    "zticks",
])

BUILTIN_CLASSES = frozenset([
    "COM",
    "CodeCompatibilityAnalysis",
    "DataTipUtilities",
    "DelaunayTri",
    "MException",
    "ParallelException",
    "RandStream",
    "SimTimeseries",
    "Tiff",
    "TriRep",
    "TriScatteredInterp",
    "VideoReader",
    "VideoWriter",
    "alphaShape",
    "arduino",
    "arrowline",
    "audioplayer",
    "audiorecorder",
    "axischild",
    "axisobj",
    "axistext",
    "calendarDuration",
    "categorical",
    "cdfepoch",
    "cell",
    "datetime",
    "decomposition",
    "delaunayTriangulation",
    "digraph",
    "duration",
    "dynamicprops",
    "editline",
    "editrect",
    "fighandle",
    "figobj",
    "ftp",
    "graph",
    "graphics",
    "griddedInterpolant",
    "handle",
    "hdf5",
    "hdftool",
    "hds",
    "hgbin",
    "inline",
    "inputParser",
    "internalArrayUtil",
    "memmap_data_handle_holder",
    "memmapfile",
    "missing",
    "mtree",
    "mupad",
    "onCleanup",
    "opaque",
    "polyshape",
    "printtemplate",
    "scatteredInterpolant",
    "scribe",
    "scribehandle",
    "scribehgobj",
    "serial",
    "string",
    "system",
    "table",
    "tabular",
    "tall",
    "tcpclient",
    "timer",
    "timerange",
    "timeseries",
    "timetable",
    "triangulation",
    "tscollection",
    "tsdata",
    "uitools",
    "vartype",
    "weboptions",
    "withtol",
])

BUILTIN_NAMESPACES = frozenset([
    "H5",
    "H5A",
    "H5D",
    "H5DS",
    "H5E",
    "H5F",
    "H5G",
    "H5I",
    "H5L",
    "H5ML",
    "H5O",
    "H5P",
    "H5R",
    "H5S",
    "H5T",
    "H5Z",
    "NET",
    "appcreate",
    "appdesigner",
    "appinstall",
    "audiovideo",
    "brushing",
    "cdflib",
    "cftool",
    "codegen",
    "connector",
    "containers",
    "datamanager",
    "dsp",
    "event",
    "exampleUtils",
    "help",
    "helpUtils",
    "images",
    "inspector",
    "internal",
    "matlab",
    "message",
    "meta",
    "mex",
    "mex_legacy",
    "mls",
    "mtree",
    "netcdf",
    "nnet",
    "objutil",
    "peermodel",
    "python",
    "spf",
    "thingspeak",
    "tsdata",
    "tsguis",
    "vision",
    "viz",
    "webutils",
])

SPECIAL_NAMES = frozenset([
    "Contents",
])



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2020-2022, Florian Schanda                    ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify it under the       ##
##  terms of the GNU General Public License as published by the Free        ##
##  Software Foundation, either version 3 of the License, or (at your       ##
##  option) any later version.                                              ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU General Public License for more details.                            ##
##                                                                          ##
##  You should have received a copy of the GNU General Public License       ##
##  along with MISS_HIT. If not, see <http://www.gnu.org/licenses/>.        ##
##                                                                          ##
##############################################################################

# pylint: disable=consider-using-dict-items

from copy import deepcopy
from miss_hit_core.m_language import MATLAB_Latest_Language


##############################################################################
# Configuration class. This will be passed to every bit of code that
# analyzes anything.
##############################################################################

class Config:
    def __init__(self, other=None):
        if other is None:
            # Set up default configuration

            self.enabled  = True
            self.language = MATLAB_Latest_Language()
            self.pragmas  = True
            # MH is enabled, language is the latest matlab, and we
            # process pragmas

            self.ignore_pragmas_with_tickets = False
            # By default, all justifications are processed

            self.style_rules = set(STYLE_RULES)
            # All rules are enabled

            self.style_config = {
                cfg_name: deepcopy(STYLE_CONFIGURATION[cfg_name].default)
                for cfg_name in STYLE_CONFIGURATION
            }
            # All rule configuration is default

            self.enabled_metrics = set(METRICS)
            # All metrics are enabled

            self.metric_limits = {}
            # No metrics have configured limits

        else:
            assert isinstance(other, Config)
            # Inherit from existing configuration
            self.enabled                     = other.enabled
            self.language                    = other.language
            self.pragmas                     = other.pragmas
            self.ignore_pragmas_with_tickets = \
                other.ignore_pragmas_with_tickets

            self.style_rules     = deepcopy(other.style_rules)
            self.style_config    = deepcopy(other.style_config)
            self.enabled_metrics = deepcopy(other.enabled_metrics)
            self.metric_limits   = deepcopy(other.metric_limits)

    def dump(self, indent=0):
        items = ["MH Configuration object"]
        items.append("Enabled  = %s" % self.enabled)
        items.append("Language = %s" % self.language)
        items.append("Pragmas  = %s" % self.pragmas)
        items.append("Style Rules")
        maxlen = max(len(item) for item in STYLE_RULES)
        for rule in sorted(STYLE_RULES):
            items.append("  %-*s = %s" % (maxlen,
                                          rule,
                                          rule in self.style_rules))
        items.append("Style Configuration")
        maxlen = max(len(item) for item in STYLE_CONFIGURATION)
        for config in sorted(STYLE_CONFIGURATION):
            items.append("  %-*s = %s" % (maxlen,
                                          config,
                                          self.style_config[config]))
        items.append("Metrics")
        maxlen = max(len(item) for item in METRICS)
        for metric in sorted(METRICS):
            if metric in self.metric_limits:
                assert metric in self.enabled_metrics
                items.append("  %-*s = limit %u" %
                             (maxlen,
                              metric,
                              self.metric_limits[metric]))
            else:
                items.append(
                    "  %-*s = %s" %
                    (maxlen,
                     metric,
                     {False: "ignore",
                      True: "report"}[metric in self.enabled_metrics]))

        for i, item in enumerate(items):
            if i == 0:
                print("%s%s" % (" " * indent, item))
            else:
                print("%s| %s" % (" " * indent, item))

    def active(self, rule):
        """ Returns true if the given rule is active. """
        assert isinstance(rule, str)
        assert rule in STYLE_RULES

        return rule in self.style_rules

    def metric_enabled(self, metric):
        """ Returns true if the given metric should be collected. """
        assert isinstance(metric, str)
        assert metric in METRICS

        return metric in self.enabled_metrics

    def metric_check(self, metric):
        """ Returns true if the given metric should be checked
            against some limit.
        """
        assert isinstance(metric, str)
        assert metric in METRICS

        return metric in self.metric_limits

    def metric_upper_limit(self, metric):
        assert isinstance(metric, str)
        assert metric in METRICS

        return self.metric_limits.get(metric, None)


##############################################################################
# Internal classes describing the kind of configuration you can put in
# a config file.
##############################################################################

class Style_Configuration:
    def __init__(self, description, default=None):
        assert isinstance(description, str)
        self.description = description
        self.default     = default


class Boolean_Style_Configuration(Style_Configuration):
    pass


class Integer_Style_Configuration(Style_Configuration):
    def __init__(self, description,
                 lower_limit=None, upper_limit=None, default=None):
        super().__init__(description, default)
        assert isinstance(lower_limit, int) or lower_limit is None
        assert isinstance(upper_limit, int) or upper_limit is None
        assert isinstance(default, int) or default is None
        assert lower_limit is None or upper_limit is None or \
            lower_limit <= upper_limit
        assert lower_limit is None or default is None or default >= lower_limit
        assert upper_limit is None or default is None or default <= upper_limit

        self.lower_limit = lower_limit
        self.upper_limit = upper_limit


class String_Style_Configuration(Style_Configuration):
    def __init__(self, description, default=None):
        assert isinstance(default, str) or default is None
        super().__init__(description, default)


class Choice_Style_Configuration(String_Style_Configuration):
    def __init__(self, description, choices):
        assert isinstance(choices, list)
        assert len(choices) >= 1
        for choice in choices:
            assert isinstance(choice, str)
        super().__init__(description, choices[0])

        self.choices = choices


class Regex_Style_Configuration(String_Style_Configuration):
    pass


class Copyright_Regex_Style_Configuration(Regex_Style_Configuration):
    pass


class Encoding_Style_Configuration(String_Style_Configuration):
    pass


class Set_Style_Configuration(Style_Configuration):
    def __init__(self, description):
        super().__init__(description, set())


class Style_Rule:
    def __init__(self, description, configuration=None):
        assert isinstance(description, str)
        assert configuration is None or \
            isinstance(configuration, dict)

        self.description   = description
        self.configuration = {}

        if configuration:
            for cfg_name in configuration:
                cfg_item = configuration[cfg_name]
                assert isinstance(cfg_name, str)
                assert isinstance(cfg_item, Style_Configuration)
                self.configuration[cfg_name] = cfg_item


class Code_Metric:
    def __init__(self, longname, description):
        assert isinstance(longname, str)
        assert isinstance(description, str)
        self.longname    = longname
        self.description = description


class File_Metric(Code_Metric):
    pass


class Function_Metric(Code_Metric):
    pass


##############################################################################
# The actual configuration options
##############################################################################

DEFAULT_NAMING_SCHEME = "([A-Z]+|[A-Z][a-z]*)(_([A-Z]+|[A-Z][a-z]*|[0-9]+))*"
# Underscore-separated acronyms or capitalised words. For example
# "Kitten_Class" or "LASER", but not "potatoFarmer".

DEFAULT_LC_NAMING_SCHEME = "[a-z]+(_[a-z]+)*"
# A much simpler naming scheme: underscore separated lowercase without
# numbers.

STYLE_RULES = {
    "file_length" : Style_Rule(
        "Ensures files do not get too big.",
        {"file_length": Integer_Style_Configuration(
            "Maximum number of allowed lines.",
            lower_limit = 1,
            default = 1000)}),

    "line_length" : Style_Rule(
        "Ensures lines do not get too long.",
        {"line_length": Integer_Style_Configuration(
            "Maximum allowed characters in any line.",
            lower_limit = 1,
            default = 80)}),

    "copyright_notice" : Style_Rule(
        "Ensures the first thing in each file is a copyright notice.",
        {
            "copyright_location": Choice_Style_Configuration(
                "Location of copyright statements",
                ["docstring", "file_header"]),
            "copyright_regex": Copyright_Regex_Style_Configuration(
                ("Regex for picking out copyright notice. Must include "
                 "named groups: 'copy', 'ystart', 'yend', and 'org'"),
                default =
                (r"(?P<copy>(Copyright \([cC]\))|((\([cC]\) )?Copyright)) "
                 r"((?P<ystart>\d\d\d\d)(-| - ))?(?P<yend>\d\d\d\d)"
                 r"( by)? *(?P<org>.*)")),
            "copyright_entity": Set_Style_Configuration(
                "Valid copyright holder."),
            "copyright_primary_entity": String_Style_Configuration(
                "Valid primary copyright holder."),
            "copyright_3rd_party_entity": Set_Style_Configuration(
                "Recognised 3rd-party copyright holder."),
            "copyright_in_embedded_code": Boolean_Style_Configuration(
                ("Enforce copyright statements in code embedded in Simulink"
                 " models."),
                False),
        }),

    "no_starting_newline" : Style_Rule(
        "Ensures files do not start with newlines."),

    "whitespace_comma" : Style_Rule(
        "Ensures there is no whitespace before a comma and whitespace (or a"
        " newline) after."),

    "whitespace_semicolon" : Style_Rule(
        "Ensures there is no whitespace before a semicolon and whitespace"
        " (or a newline) after."),

    "spurious_row_comma" : Style_Rule(
        "Ensures there are no unnecessary leading or trailing commas in a"
        " matrix or cell."),

    "spurious_row_semicolon" : Style_Rule(
        "Ensures there are no unnecessary leading or trailing semicolons in a"
        " matrix or cell."),

    "whitespace_colon" : Style_Rule(
        "Ensures there is no whitespace around colons except if they come"
        " after a comma."),

    "whitespace_assignment" : Style_Rule(
        "Ensures there is whitespace around the assignment operator (=)."),

    "whitespace_brackets" : Style_Rule(
        "Ensures no whitespace after (/[, and no whitespace before )/]."),

    "whitespace_keywords" : Style_Rule(
        "Ensures whitespace after some words, such as if, or properties."),

    "whitespace_comments" : Style_Rule(
        "Ensures whitespace before comments and whitespace between the % and"
        " the body of the comment. Pragmas (%#) are exempt."),

    "whitespace_continuation" : Style_Rule(
        "Ensures whitespace before continuations and whitespace between"
        " the ... and any in-line comment."),

    "whitespace_around_functions" : Style_Rule(
        "Ensures there is whitespace around functions."),

    "operator_after_continuation" : Style_Rule(
        "Complains about operators after a line continuation."),

    "dangerous_continuation": Style_Rule(
        "Flag misleading line continuations."),

    "useless_continuation" : Style_Rule(
        "Flag unnecessary line continuations."),

    "operator_whitespace" : Style_Rule(
        "Enfore whitespace around unary and binary operators."),

    "end_of_statements" : Style_Rule(
        "Ensures consistent ending of statements."),

    "builtin_shadow" : Style_Rule(
        "Checks that assignments do not overwrite builtin functions such as"
        " true, false, or pi."),

    "naming_scripts" : Style_Rule(
        "Checks names of script files.",
        {
            "regex_script_name" : Regex_Style_Configuration(
                "Regex for script names",
                default = DEFAULT_NAMING_SCHEME),
        }),

    "naming_functions" : Style_Rule(
        "Checks names of functions, nested functions, and class methods.",
        {
            "regex_function_name" : Regex_Style_Configuration(
                "Regex for function names",
                default = DEFAULT_NAMING_SCHEME),
            "regex_nested_name" : Regex_Style_Configuration(
                "Regex for nested function names",
                default = DEFAULT_NAMING_SCHEME),
            "regex_method_name" : Regex_Style_Configuration(
                "Regex for class method names",
                default = DEFAULT_LC_NAMING_SCHEME)
        }),

    "naming_parameters" : Style_Rule(
        "Checks names of function and method parameters.",
        {
            "regex_parameter_name" : Regex_Style_Configuration(
                "Regex for parameter names",
                default = DEFAULT_LC_NAMING_SCHEME)
        }),

    "naming_classes" : Style_Rule(
        "Checks names of classes.",
        {
            "regex_class_name" : Regex_Style_Configuration(
                "Regex for class names",
                default = DEFAULT_NAMING_SCHEME)
        }),

    "naming_enumerations" : Style_Rule(
        "Checks names of enumerations.",
        {
            "regex_enumeration_name" : Regex_Style_Configuration(
                "Regex for parameter names",
                default = DEFAULT_NAMING_SCHEME)
        }),

    "indentation" : Style_Rule(
        "Make indentation consistent.",
        {
            "tab_width": Integer_Style_Configuration(
                "Number of spaces to indent by.",
                lower_limit = 2,
                default = 4),
            "indent_function_file_body" : Boolean_Style_Configuration(
                "Indent the body of functions in a function file.",
                default = True),
            "align_round_brackets" : Boolean_Style_Configuration(
                "Align continuations inside normal brackets.",
                default = True),
            "align_other_brackets" : Boolean_Style_Configuration(
                "Align continuations inside matrix and cell expressions.",
                default = True)
        }),

    "force_newlines" : Style_Rule(
        "Make newlines consistent.",
        {
            "newline_style" : Choice_Style_Configuration(
                "Newline style",
                ["native", "crlf", "lf", "cr"]),
        }),

    "redundant_brackets" : Style_Rule(
        "Check for obviously useless brackets. Does not complain about"
        " brackets added for clarifying operator precedence."),

    "annotation_whitespace" : Style_Rule("Check for a space after %|"),

    "implicit_shortcircuit": Style_Rule(
        "Complain about implicit short-circuit operations in"
        " if or while guards"),

    "unicode": Style_Rule(
        "Complain about non-conforming characters in source files",
        {
            "enforce_encoding": Encoding_Style_Configuration(
                "Override the encoding to enforence, by default ASCII",
                default = "ascii"),
            "enforce_encoding_comments" : Boolean_Style_Configuration(
                ("Also check comments/continuations for conforming to "
                 "the encoding"),
                default = True)
        }),
}

STYLE_CONFIGURATION = {
    cfg_name: STYLE_RULES[rule_name].configuration[cfg_name]
    for rule_name in STYLE_RULES
    for cfg_name in STYLE_RULES[rule_name].configuration
}
# A flat list of all style configuration items

STYLE_CONFIGURATION["regex_tickets"] = Regex_Style_Configuration(
    "Regex for tickets in your issue tracking system",
    default = "")
# Add the special regex that identifies tickets (not part of any
# rule).

METRICS = {
    "file_length" : File_Metric(
        "File lines",
        "Number of lines in each file."),

    "function_length" : Function_Metric(
        "Function lines",
        "Number for lines for each function."),

    "npath" : Function_Metric(
        "Number of paths",
        "Approximation for maximum number of paths in function."),

    "cnest" : Function_Metric(
        "Control nesting",
        "Maximum nesting of control structures."),

    "parameters" : Function_Metric(
        "Parameters",
        "Number of input/output parameters."),

    "globals" : Function_Metric(
        "Globals",
        "Number of direct (non-transitive) global variables."),

    "persistent" : Function_Metric(
        "Persistents",
        "Number of persistent variables."),

    "cyc" : Function_Metric(
        "Cyclomatic complexity",
        "The McCabe cyclomatic complexity metric."),
}
FILE_METRICS = sorted(metric
                      for metric in METRICS
                      if isinstance(METRICS[metric], File_Metric))
FUNCTION_METRICS = sorted(metric
                          for metric in METRICS
                          if isinstance(METRICS[metric], Function_Metric))
assert len(FILE_METRICS) + len(FUNCTION_METRICS) == len(METRICS)



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2019-2024, Florian Schanda                    ##
##              Copyright (C) 2019, Zenuity AB                              ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify it under the       ##
##  terms of the GNU General Public License as published by the Free        ##
##  Software Foundation, either version 3 of the License, or (at your       ##
##  option) any later version.                                              ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU General Public License for more details.                            ##
##                                                                          ##
##  You should have received a copy of the GNU General Public License       ##
##  along with MISS_HIT. If not, see <http://www.gnu.org/licenses/>.        ##
##                                                                          ##
##############################################################################

import os
import sys
import html
import json

from miss_hit_core import pathutil
from miss_hit_core.config import STYLE_RULES, METRICS


class Location:
    """ This fully describes where a message originates from.

    * filename must be an actual file where the message is attached
    * blockname can describe a virtual construct (e.g. simulink block)
    * line is the line number (starts at 1)
    * col_start and col_end describe the column (starts at 0)
    * context is a replication of the line that contains the offending
      construct
    """
    def __init__(self,
                 filename,
                 line=None,
                 col_start=None,
                 col_end=None,
                 context=None,
                 blockname=None):
        assert isinstance(filename, str)
        assert blockname is None or isinstance(blockname, str)
        assert line is None or (isinstance(line, int) and line >= 1)
        assert col_start is None or (isinstance(col_start, int) and
                                     col_start >= 0)
        assert col_end is None or (isinstance(col_end, int) and
                                   col_end >= 0 and
                                   col_start is not None)
        assert context is None or isinstance(context, str)

        self.filename = filename.replace("\\", "/")
        # We canonicalise filenames so that windows and linux produce
        # the same output.

        self.blockname = blockname

        self.line = line
        self.col_start = col_start
        if col_end is None:
            self.col_end = col_start
        else:
            self.col_end = max(col_start, col_end)
        self.context = context

    def __str__(self):
        return "Location(%s,l=%s,b=%s)" % (self.filename,
                                           self.line,
                                           self.blockname)

    def __lt__(self, other):
        assert isinstance(other, Location)

        return (self.filename,
                self.line if self.line else 0,
                self.col_start if self.col_start else -1) < \
            (other.filename,
             other.line if other.line else 0,
             other.col_start if other.col_start else -1)

    def to_json(self, detailed=True):
        rv = {"filename": self.filename}
        if self.blockname:
            rv["block"] = self.blockname
        if self.line:
            rv["line"] = self.line
        if self.col_start:
            rv["col_start"] = self.col_start
        if detailed and self.col_end:
            rv["col_end"] = self.col_end
        if detailed and self.context:
            rv["context"] = self.context
        return rv

    def short_string(self):
        rv = self.filename
        if self.line:
            rv += ":%u" % self.line
        return rv


class ICE(Exception):
    """ Internal compiler errors """
    def __init__(self, reason):
        super().__init__(reason)
        self.reason = reason


class Error(Exception):
    """ Any other, possibly recoverable, errors """
    def __init__(self, location, message):
        assert isinstance(location, Location)
        assert isinstance(message, str)

        super().__init__(message)
        self.location = location
        self.message = message


class Justification:
    def __init__(self, token):
        # assert isinstance(token, m_lexer.MATLAB_Token)
        assert token.kind in ("COMMENT", "CONTINUATION")

        self.token = token
        self.used = False


class Style_Justification(Justification):
    pass


class Message:
    def __init__(self, location, kind, check_id, message, fatal, autofixed):
        assert isinstance(location, Location)
        assert kind in ("info",       # diagnostics and information
                        "style",      # style issues (from mh_style)
                        "metric",     # code metrics (from mh_metric)
                        "check",      # defects (from mh_lint and mh_prove)
                        "warning",    # other issues
                        "lex error",  # errors in the lexing phase
                        "error")      # other errors
        if check_id is not None:
            assert isinstance(check_id, str)
            if kind == "style":
                assert check_id in STYLE_RULES
            elif kind == "metric":
                assert check_id in METRICS
        assert isinstance(message, str)
        assert isinstance(fatal, bool)
        assert isinstance(autofixed, bool)
        assert not fatal or kind in ("lex error", "error"), \
            "fatal=%s, kind=%s violates precondition" % (fatal, kind)

        self.location  = location
        self.kind      = kind
        self.check_id  = check_id
        self.severity  = "medium"  # can be low, medium, high
        self.message   = message
        self.fixed     = autofixed
        self.fatal     = fatal
        self.justified = False

    def __str__(self):
        return "Message(%s,%s,%s)" % (self.location,
                                      self.kind,
                                      repr(self.message))

    def __lt__(self, other):
        assert isinstance(other, Message)

        return self.location < other.location

    def check_justification(self, justification):
        if self.kind == "style" and \
           isinstance(justification, Style_Justification):
            if self.location.line == justification.token.location.line:
                self.justified = True
                justification.used = True

    def to_json(self):
        return {"location"  : self.location.to_json(),
                "kind"      : self.kind,
                "severity"  : self.severity,
                "message"   : self.message,
                "fixable"   : self.fixed,
                "fatal"     : self.fatal}


class Check_Message(Message):
    def __init__(self, location, severity, check_id, message):
        super().__init__(location  = location,
                         kind      = "check",
                         check_id  = check_id,
                         message   = message,
                         fatal     = False,
                         autofixed = False)
        assert severity in ("low", "medium", "high")
        self.severity = severity


class Message_Handler:
    """ All messages should be routed through this class """
    def __init__(self, tool_id):
        assert tool_id in ("debug",
                           "style", "metric",
                           "lint", "trace", "bmc",
                           "diff", "copyright")

        self.tool_id = tool_id

        self.style_issues = 0
        self.metric_issues = 0
        self.metric_justifications = 0
        self.checks = 0
        self.warnings = 0
        self.errors = 0
        self.justified = 0
        self.files = set()
        self.excluded_files = set()
        self.seen_files = set()

        self.autofix = False
        self.colour = False
        self.show_context = True
        self.show_style = True
        self.show_checks = False
        self.sort_messages = True

        self.messages = {}        # file -> line -> [message]
        self.justifications = {}  # file -> line -> [justification]

    def debug_dump(self):
        print("Debug dump for Message_Handler object")
        for file_name in sorted(self.messages):
            print("> File name: %s" % file_name)
            for line in sorted(self.messages[file_name]):
                print("  Line: %u" % line)
                for msg in self.messages[file_name][line]:
                    print("    ", str(msg))

    def reset_seen(self):
        self.seen_files = set()

    def fork(self):
        rv = Message_Handler(self.tool_id)
        self.fork_copy_attributes(rv)
        return rv

    def fork_copy_attributes(self, other):
        other.autofix       = self.autofix
        other.colour        = self.colour
        other.show_context  = self.show_context
        other.show_style    = self.show_style
        other.show_checks   = self.show_checks
        other.sort_messages = self.sort_messages

    def integrate(self, other):
        assert isinstance(other, Message_Handler)
        assert self.autofix       == other.autofix
        assert self.colour        == other.colour
        assert self.show_context  == other.show_context
        assert self.show_style    == other.show_style
        assert self.sort_messages == other.sort_messages

        self.style_issues          += other.style_issues
        self.metric_issues         += other.metric_issues
        self.metric_justifications += other.metric_justifications
        self.checks                += other.checks
        self.warnings              += other.warnings
        self.errors                += other.errors
        self.justified             += other.justified
        self.files                 |= other.files
        self.excluded_files        |= other.excluded_files
        self.seen_files            |= other.seen_files

        for filename in other.messages:
            if filename not in self.messages:
                self.messages[filename] = {}
            for line in other.messages[filename]:
                if line not in self.messages[filename]:
                    self.messages[filename][line] = []
                self.messages[filename][line] += \
                  other.messages[filename][line]

        for filename in other.justifications:
            if filename not in self.justifications:
                self.justifications[filename] = {}
            for line in other.justifications[filename]:
                if line not in self.justifications[filename]:
                    self.justifications[filename][line] = []
                self.justifications[filename][line] += \
                  other.justifications[filename][line]

    def register_file(self, filename):
        assert isinstance(filename, str)
        canonical_filename = filename.replace("\\", "/")
        assert canonical_filename not in self.files
        assert canonical_filename not in self.excluded_files

        self.files.add(canonical_filename)
        self.seen_files.add(canonical_filename)
        self.messages[canonical_filename] = {}
        self.justifications[canonical_filename] = {}

    def register_exclusion(self, filename):
        assert isinstance(filename, str)
        canonical_filename = filename.replace("\\", "/")
        assert canonical_filename not in self.files
        assert canonical_filename not in self.excluded_files

        self.excluded_files.add(canonical_filename)

    def unregister_file(self, filename):
        assert isinstance(filename, str)
        canonical_filename = filename.replace("\\", "/")
        assert canonical_filename in self.files

        self.files.remove(canonical_filename)
        del self.messages[canonical_filename]
        del self.justifications[canonical_filename]

    def process_message(self, message):
        # Count the message
        if message.justified:
            self.justified += 1
            return
        elif message.kind == "info":
            pass
        elif message.kind == "style":
            if self.show_style:
                self.style_issues += 1
            else:
                return
        elif message.kind == "metric":
            self.metric_issues += 1
        elif message.kind == "check":
            if self.show_checks:
                self.checks += 1
            else:
                return
        elif message.kind == "warning":
            self.warnings += 1
        elif message.kind in ("lex error", "error"):
            self.errors += 1
        else:
            raise ICE("unexpeced message kind %s" % message.kind)

        # Emit
        self.emit_message(message)

    def emit_message(self, message):
        if self.colour:
            if message.kind in ("error", "lex error"):
                kstring = "\033[31;1m%s\033[0m" % message.kind
            elif message.kind == "warning":
                kstring = "\033[31m%s\033[0m" % message.kind
            elif message.kind == "check":
                if message.severity == "low":
                    kstring = "%s (\033[34m%s\033[0m)" % (message.kind,
                                                          message.severity)
                elif message.severity == "high":
                    kstring = "%s (\033[31;1m%s\033[0m)" % (message.kind,
                                                            message.severity)
                else:
                    kstring = "%s (\033[33m%s\033[0m)" % (message.kind,
                                                          message.severity)
            else:
                kstring = message.kind
        elif message.kind == "check":
            kstring = message.kind + " (" + message.severity + ")"
        else:
            kstring = message.kind

        mtext = message.message

        if message.fixed and self.autofix:
            mtext += " [fixed]"

        if message.check_id:
            mtext += " [%s]" % message.check_id

        if message.location.blockname is None:
            full_location = message.location.filename
        else:
            full_location = "%s/%s" % (message.location.filename,
                                       message.location.blockname)

        if message.location.line is None:
            print("%s: %s: %s" % (full_location,
                                  kstring,
                                  mtext))
        elif message.location.col_start is None:
            print("%s:%u: %s: %s" % (full_location,
                                     message.location.line,
                                     kstring,
                                     mtext))
        else:
            if message.location.context is None:
                show_context = False
            elif len(message.location.context.strip()) > 0:
                show_context = self.show_context
            else:
                show_context = False

            if show_context:
                print("In %s, line %u" % (full_location,
                                          message.location.line))
                print("| " + message.location.context.replace("\t", " "))
                print("| " +
                      (" " * message.location.col_start) +
                      ("^" * (message.location.col_end -
                              message.location.col_start + 1)) +
                      " %s: %s" % (kstring, mtext))
            else:
                print("%s:%u:%u: %s: %s" % (full_location,
                                            message.location.line,
                                            message.location.col_start,
                                            kstring,
                                            mtext))

    def emit_summary(self):
        tmp = "MISS_HIT %s Summary: " % self.tool_id.capitalize()
        stats = ["%u file(s) analysed" % len(self.seen_files)]
        if self.style_issues:
            stats.append("%u style issue(s)" % self.style_issues)
        if self.metric_issues:
            stats.append("%u metric deviations(s)" % self.metric_issues)
        if self.checks:
            stats.append("%u check(s)" % self.checks)
        if self.warnings:
            stats.append("%u warning(s)" % self.warnings)
        if self.errors:
            stats.append("%u error(s)" % self.errors)
        if len(stats) == 1:
            stats.append("everything seems fine")
        if self.justified > 0:
            stats.append("%u justified message(s)" % self.justified)
        if self.metric_justifications > 0:
            stats.append("%u justified metric deviations(s)" %
                         self.metric_justifications)
        tmp += ", ".join(stats)
        if self.excluded_files:
            tmp += ("; %u file(s) excluded from analysis" %
                    len(self.excluded_files))
        print(tmp)

    def register_message(self, msg):
        assert isinstance(msg, Message)

        if msg.location.filename not in self.files:
            raise ICE("attempted to emit message on unknown file '%s'" %
                      msg.location.filename)

        if self.sort_messages:
            # Add message to list
            messages = self.messages[msg.location.filename]
            if msg.location.line not in messages:
                messages[msg.location.line] = [msg]
            else:
                messages[msg.location.line].append(msg)

            # Check if a justification applies
            just = self.messages[msg.location.filename]
            if msg.location.line in just:
                for j in just[msg.location.line]:
                    msg.check_justification(j)

        else:
            # Otherwise just immediately emit it
            self.process_message(msg)

        # Raise exception for fatal messages
        if msg.fatal:
            raise Error(msg.location, msg.message)

    def register_justification(self, token):
        # assert isinstance(token, m_lexer.MATLAB_Token)
        assert token.kind in ("COMMENT", "CONTINUATION")

        if token.location.filename not in self.files:
            raise ICE("attempted to add justification to an unknown file")

        if not self.sort_messages:
            return

        if "mh:ignore_style" in token.value:
            justification = Style_Justification(token)
        else:
            self.warning(token.location,
                         "invalid justification not recognized")
            return

        # Add justification
        just = self.justifications[token.location.filename]
        if token.location.line not in just:
            just[token.location.line] = [justification]
        else:
            just[token.location.line].append(justification)

        # Check if a message is justified
        messages = self.messages[token.location.filename]
        if token.location.line in messages:
            for msg in messages[token.location.line]:
                msg.check_justification(justification)

    def info(self, location, message):
        msg = Message(location  = location,
                      kind      = "info",
                      check_id  = None,
                      message   = message,
                      fatal     = False,
                      autofixed = False)
        self.register_message(msg)

    def style_issue(self, location, message, check_id, autofix=False):
        msg = Message(location  = location,
                      kind      = "style",
                      check_id  = check_id,
                      message   = message,
                      fatal     = False,
                      autofixed = autofix)
        self.register_message(msg)

    def metric_issue(self, location, message, metric_id):
        msg = Message(location  = location,
                      kind      = "metric",
                      check_id  = metric_id,
                      message   = message,
                      fatal     = False,
                      autofixed = False)
        self.register_message(msg)

    def check(self, location, message, check_id, severity="medium"):
        msg = Check_Message(location = location,
                            severity = severity,
                            check_id = check_id,
                            message  = message)
        self.register_message(msg)

    def warning(self, location, message):
        msg = Message(location  = location,
                      kind      = "warning",
                      check_id  = None,
                      message   = message,
                      fatal     = False,
                      autofixed = False)
        self.register_message(msg)

    def lex_error(self, location, message, fatal=True):
        msg = Message(location  = location,
                      kind      = "lex error",
                      check_id  = None,
                      message   = message,
                      fatal     = fatal,
                      autofixed = False)
        self.register_message(msg)

    def error(self, location, message, fatal=True):
        msg = Message(location  = location,
                      kind      = "error",
                      check_id  = None,
                      message   = message,
                      fatal     = fatal,
                      autofixed = False)
        self.register_message(msg)

    def config_error(self, location, message):
        # This is for raising errors in config files _only_.
        canonical_filename = location.filename.replace("\\", "/")
        if canonical_filename not in self.files:
            self.register_file(location.filename)
        self.error(location, message)

    def finalize_file(self, filename):
        assert isinstance(filename, str)
        canonical_filename = filename.replace("\\", "/")
        assert canonical_filename in self.files

        # New messages for justifications that did not apply
        for justifications in self.justifications[canonical_filename].values():
            for justification in justifications:
                if not justification.used:
                    self.warning(justification.token.location,
                                 "style justification does not apply")

        # Process messages
        for messages in sorted(self.messages[canonical_filename].values()):
            for message in sorted(messages):
                self.process_message(message)

        # Remove file
        self.unregister_file(canonical_filename)

    def command_line_error(self, message):
        print("%s: error: %s" % (os.path.basename(sys.argv[0]), message))
        sys.exit(1)

    def summary_and_exit(self):
        files = list(self.messages)
        for filename in files:
            self.finalize_file(filename)

        self.emit_summary()

        if self.style_issues or \
           self.metric_issues or \
           self.warnings or \
           self.errors:
            sys.exit(1)
        else:
            sys.exit(0)


class File_Based_Message_Handler(Message_Handler):
    def __init__(self, tool_id, filename):
        super().__init__(tool_id)
        self.filename = filename
        self.fd       = None

    def fork(self):
        raise ICE("unimplemented abstract class")

    def setup_fd(self):
        raise ICE("unimplemented abstract class")


class HTML_Message_Handler(File_Based_Message_Handler):
    def __init__(self, tool_id, filename):
        super().__init__(tool_id, filename)
        self.last_file = None

    def fork(self):
        rv = HTML_Message_Handler(self.tool_id, self.filename)
        self.fork_copy_attributes(rv)
        return rv

    def setup_fd(self):
        # pylint: disable=consider-using-with
        if self.fd is not None:
            return

        self.fd = open(self.filename, "w", encoding="UTF-8")
        self.fd.write("<!DOCTYPE html>\n")
        self.fd.write("<html>\n")
        self.fd.write("<head>\n")
        self.fd.write("<meta charset=\"UTF-8\">\n")
        # Link style-sheet with a relative path based on where the
        # output report file will be
        self.fd.write("<link rel=\"stylesheet\" href=\"file:%s\">\n" %
                      os.path.relpath(os.path.join(sys.path[0],
                                                   "docs",
                                                   "style.css"),
                                      os.path.dirname(
                                          pathutil.abspath(self.filename))).
                      replace("\\", "/"))
        self.fd.write("<title>MISS_HIT Report</title>\n")
        self.fd.write("</head>\n")
        self.fd.write("<body>\n")
        self.fd.write("<header>MISS_HIT Report</header>\n")
        self.fd.write("<main>\n")
        self.fd.write("<div></div>\n")
        self.fd.write("<h1>Issues identified</h1>\n")
        self.fd.write("<section>\n")

    def emit_message(self, message):
        self.setup_fd()

        if self.last_file != message.location.filename:
            self.last_file = message.location.filename
            self.fd.write("<h2>%s</h2>\n" % message.location.filename)

        mtext = message.message
        if message.fixed and self.autofix:
            mtext += " [fixed]"

        self.fd.write("<div class=\"message\">")
        if message.location.col_start:
            self.fd.write("<a href=\"matlab:opentoline('%s', %u, %u)\">" %
                          (message.location.filename,
                           message.location.line,
                           message.location.col_start + 1))
            self.fd.write("%s: line %u:" % (message.location.filename,
                                            message.location.line))
        elif message.location.line:
            self.fd.write("<a href=\"matlab:opentoline('%s', %u)\">" %
                          (message.location.filename,
                           message.location.line))
            self.fd.write("%s: line %u:" % (message.location.filename,
                                            message.location.line))
        else:
            self.fd.write("<a href=\"matlab:opentoline('%s')\">" %
                          message.location.filename)
            self.fd.write("%s:" % message.location.filename)
        self.fd.write("</a>")

        if message.kind == "check":
            self.fd.write(" %s (%s):" % (message.kind,
                                         message.severity))
        else:
            self.fd.write(" %s:" % message.kind)

        self.fd.write(" %s" % html.escape(message.message))

        self.fd.write("</div>\n")

    def emit_summary(self):
        self.setup_fd()
        super().emit_summary()
        if not (self.style_issues or self.warnings or self.errors):
            self.fd.write("<div>Everything is fine :)</div>")
        self.fd.write("</section>\n")
        self.fd.write("</main>\n")
        self.fd.write("</body>\n")
        self.fd.write("</html>\n")
        self.fd.close()
        self.fd = None


class JSON_Message_Handler(File_Based_Message_Handler):
    def __init__(self, tool_id, filename):
        super().__init__(tool_id, filename)
        self.blob = None

    def fork(self):
        rv = JSON_Message_Handler(self.tool_id, self.filename)
        self.fork_copy_attributes(rv)
        return rv

    def setup_fd(self):
        # pylint: disable=consider-using-with
        if self.fd is not None:
            return

        self.fd = open(self.filename, "w", encoding="UTF-8")
        self.blob = {}

    def emit_message(self, message):
        self.setup_fd()

        if message.location.filename not in self.blob:
            self.blob[message.location.filename] = []

        self.blob[message.location.filename].append(message.to_json())

    def emit_summary(self):
        self.setup_fd()
        super().emit_summary()
        json.dump(self.blob, self.fd, indent=2)
        self.fd.write("\n")
        self.fd.close()



#!/usr/bin/env python3
##############################################################################
##                                                                          ##
##          MATLAB Independent, Small & Safe, High Integrity Tools          ##
##                                                                          ##
##              Copyright (C) 2019-2024, Florian Schanda                    ##
##              Copyright (C) 2019-2020, Zenuity AB                         ##
##                                                                          ##
##  This file is part of MISS_HIT.                                          ##
##                                                                          ##
##  MATLAB Independent, Small & Safe, High Integrity Tools (MISS_HIT) is    ##
##  free software: you can redistribute it and/or modify it under the       ##
##  terms of the GNU General Public License as published by the Free        ##
##  Software Foundation, either version 3 of the License, or (at your       ##
##  option) any later version.                                              ##
##                                                                          ##
##  MISS_HIT is distributed in the hope that it will be useful,             ##
##  but WITHOUT ANY WARRANTY; without even the implied warranty of          ##
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           ##
##  GNU General Public License for more details.                            ##
##                                                                          ##
##  You should have received a copy of the GNU General Public License       ##
##  along with MISS_HIT. If not, see <http://www.gnu.org/licenses/>.        ##
##                                                                          ##
##############################################################################

import os

from miss_hit_core import pathutil
from miss_hit_core.config import Config, METRICS
from miss_hit_core.errors import ICE, Message_Handler
from miss_hit_core.m_ast import *
from miss_hit_core.m_language import Language
from miss_hit_core.m_lexer import Token_Generator


IGNORED_TOKENS = frozenset(["COMMENT"])

# A general note on the embedded grammar fragments: we're using Ada RM
# style grammar. This means:
#
#   ::=               indicates a production
#   [ grammar ]       means zero or one
#   { grammar }       means zero or more repeating
#   ( grammar )       groups a grammar fragment (usually a disjunction)
#   '{'               means the token {
#   |                 means an alternative production

# Operator precedence as of MATLAB 2019b
# https://www.mathworks.com/help/matlab/matlab_prog/operator-precedence.html
#
# 1. Parentheses ()
#
# 2. Transpose (.'), power (.^), complex conjugate transpose ('),
#    matrix power (^)
#
# 3. Power with unary minus (.^-), unary plus (.^+), or logical
#    negation (.^~) as well as matrix power with unary minus (^-), unary
#    plus (^+), or logical negation (^~).
#
#    Note: Although most operators work from left to right, the
#    operators (^-), (.^-), (^+), (.^+), (^~), and (.^~) work from
#    second from the right to left. It is recommended that you use
#    parentheses to explicitly specify the intended precedence of
#    statements containing these operator combinations.
#
# 4. Unary plus (+), unary minus (-), logical negation (~)
#
# 5. Multiplication (.*), right division (./), left division (.\),
#    matrix multiplication (*), matrix right division (/), matrix left
#    division (\)
#
# 6. Addition (+), subtraction (-)
#
# 7. Colon operator (:)
#
# 8. Less than (<), less than or equal to (<=), greater than (>),
#    greater than or equal to (>=), equal to (==), not equal to (~=)
#
# 9. Element-wise AND (&)
#
# 10. Element-wise OR (|)
#
# 11. Short-circuit AND (&&)
#
# 12. Short-circuit OR (||)


class MATLAB_Parser:
    def __init__(self, mh, lexer, cfg):
        assert isinstance(mh, Message_Handler)
        assert isinstance(lexer, Token_Generator)
        assert isinstance(cfg, Config)

        self.lexer = lexer
        self.language = self.lexer.language
        self.mh    = mh
        self.cfg   = cfg

        self.context = []

        self.functions_require_end = False
        # If true, we have encountered a function with end. This means
        # all of them must have an end.

        self.in_shortcircuit_context = False
        # Some places change the meaning of & and |, we use this to
        # keep track of it.
        #
        # Note: this is not always correct, in some cases & means &
        # even inside a top level if. Specifically if either operand
        # is an array (see test parser/bug_139). For now this does not
        # matter (the style rule that would make use of this has been
        # disabled) but we need to sort this out in semantic analysis.
        #
        # Curiously mlint seems to share this bug.

        # pylint: disable=invalid-name
        self.ct  = None
        self.nt  = None
        self.nnt = None
        # pylint: enable=invalid-name

        self.debug_tree = False

        self.skip()
        self.skip()

    def sc_context(self, enabled):
        assert isinstance(enabled, bool)

        class CM:
            def __init__(self, parser, current_value, new_value):
                self.parser    = parser
                self.old_value = current_value
                self.new_value = new_value

            def __enter__(self):
                self.parser.in_shortcircuit_context = self.new_value

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.parser.in_shortcircuit_context = self.old_value

        return CM(self, self.in_shortcircuit_context, enabled)

    def push_context(self, kind):
        assert kind in ("function", "classdef",
                        "loop", "if", "switch",
                        "block")
        self.context.append(kind)

    def pop_context(self):
        if self.context:
            self.context.pop()
        else:
            raise ICE("context is empty")

    def in_context(self, kind):
        assert kind in ("function", "classdef",
                        "loop", "if", "switch",
                        "block")
        for k in reversed(self.context):
            if kind == k:
                return True
            elif k in ("function", "classdef"):
                return False
        return False

    def skip(self):
        def should_skip(token):
            # Returns true iff this token should be completely passed
            # over by the parser. We do this since the lexer produces
            # tokens for whitespace and comments (needed for pretty
            # printing in mh_style).

            if not token:
                # Never skip EOF
                return False
            elif token.kind in ("COMMENT",
                                "CONTINUATION",
                                "ANNOTATION"):
                # Skip all of these in all cases
                return True
            elif token.annotation and token.kind == "NEWLINE":
                # In MATLAB newlines are important. In MISS_HIT
                # annotations they are not, so if we're in an
                # annotation we can just ignore them.
                return True
            else:
                # Otherwise don't skip
                return False

        self.ct = self.nt
        self.nt = self.nnt
        self.nnt = self.lexer.token()

        while self.nnt:
            # Skip comments, continuations and annotation indications
            while should_skip(self.nnt):
                self.nnt = self.lexer.token()

            # Join new-lines
            if (self.nnt and
                self.nt and
                self.nnt.kind == "NEWLINE" and
                self.nt.kind == "NEWLINE"):
                self.nnt = self.lexer.token()
            else:
                break

    def match(self, kind, value=None):
        assert kind in self.language.token_kinds
        self.skip()
        if self.ct is None:
            self.mh.error(self.lexer.get_file_loc(),
                          "expected %s, reached EOF instead" % kind)
        elif self.ct.annotation:
            self.mh.error(self.ct.location,
                          "expected %s, "
                          "found miss_hit annotation instead" %
                          kind)
        elif self.ct.kind != kind:
            if value:
                self.mh.error(self.ct.location,
                              "expected %s(%s), found %s instead" %
                              (kind, value, self.ct.kind))
            else:
                self.mh.error(self.ct.location,
                              "expected %s, found %s instead" % (kind,
                                                                 self.ct.kind))

        elif value and self.ct.value != value:
            self.mh.error(self.ct.location,
                          "expected %s(%s), found %s(%s) instead" %
                          (kind, value, self.ct.kind, self.ct.value))

    def amatch(self, kind, value=None):
        assert kind in self.language.token_kinds
        self.skip()
        if self.ct is None:
            self.mh.error(self.lexer.get_file_loc(),
                          "expected %s, reached EOF instead" % kind)
        elif not self.ct.annotation:
            self.mh.error(self.ct.location,
                          "expected %s annotation, "
                          "found normal program text instead" %
                          kind)
        elif self.ct.kind != kind:
            if value:
                self.mh.error(self.ct.location,
                              "expected %s(%s), found %s instead" %
                              (kind, value, self.ct.kind))
            else:
                self.mh.error(self.ct.location,
                              "expected %s, found %s instead" % (kind,
                                                                 self.ct.kind))

        elif value and self.ct.value != value:
            self.mh.error(self.ct.location,
                          "expected %s(%s), found %s(%s) instead" %
                          (kind, value, self.ct.kind, self.ct.value))

    def match_eof(self):
        self.skip()
        if self.ct is not None:
            if self.ct.annotation:
                self.mh.error(self.ct.location,
                              "expected end of file, "
                              "found annotation %s instead" %
                              self.ct.kind)
            else:
                self.mh.error(self.ct.location,
                              "expected end of file, found %s instead" %
                              self.ct.kind)

    def peek_annotation(self):
        return self.nt and self.nt.annotation

    def peek(self, kind, value=None):
        assert kind in self.language.token_kinds
        if self.nt and \
           self.nt.kind == kind and \
           not self.nt.annotation:
            if value is None:
                return True
            else:
                return self.nt.value == value
        else:
            return False

    def peek2(self, kind, value=None):
        assert kind in self.language.token_kinds
        if self.nnt and \
           self.nnt.kind == kind and \
           not self.nnt.annotation:
            if value is None:
                return True
            else:
                return self.nnt.value == value
        else:
            return False

    def apeek(self, kind, value=None):
        assert kind in self.language.token_kinds
        if self.nt and \
           self.nt.kind == kind and \
           self.nt.annotation:
            if value is None:
                return True
            else:
                return self.nt.value == value
        else:
            return False

    def peek_eof(self):
        return self.nt is None

    ##########################################################################
    # Some style-related helper functions

    def set_expression_brackets(self, n_expr, t_open, t_close):
        assert isinstance(n_expr, Expression)

        # We already have brackets, so any old ones are redundant
        self.check_redundant_brackets(n_expr)

        n_expr.set_enclosing_brackets(t_open, t_close)

    def check_redundant_brackets(self, n_expr):
        assert isinstance(n_expr, Expression), \
            "required expression, not %s" % n_expr.__class__.__name__

        if self.cfg.active("redundant_brackets") and \
           n_expr.t_bracket_open:
            self.mh.style_issue(n_expr.t_bracket_open.location,
                                "redundant parenthesis",
                                "redundant_brackets",
                                True)
            n_expr.t_bracket_open.fix.delete = True
            n_expr.t_bracket_close.fix.delete = True

    ##########################################################################
    # Parsing

    def peek_eos(self):
        return self.peek("SEMICOLON") or \
            self.peek("COMMA") or \
            self.peek("NEWLINE")

    def match_eos(self, n_ast, semi = "", allow_nothing = False):
        """Match end-of-statement

        This generally matches a newline. If semi is set to ";", then
        it matches a semicolon followed by a newline. Anything that
        deviates from this will create style issues.

        Any termination tokens are attached to the node given by
        n_ast.

        A syntax error is raised if ot statement terminator is found,
        unless allow_nothing is set to true.

        """
        assert isinstance(n_ast, Node)
        assert semi in ("", ";")
        assert isinstance(allow_nothing, bool)

        ending_token = self.ct
        # The last token of the previous thing. We might need it later
        # to attach error messages or to record autofix instructions.

        # Get all terminator tokens that follow our ending token.
        terminator_tokens = []
        first_newline = None
        while self.peek_eos():
            self.skip()
            self.ct.set_ast(n_ast)
            self.ct.fix.statement_terminator = True
            terminator_tokens.append(self.ct)
            if self.ct.kind == "NEWLINE" and first_newline is None:
                first_newline = len(terminator_tokens) - 1
                break
        while self.peek_eos():  # and not self.peek("NEWLINE"):
            self.skip()
            self.ct.set_ast(n_ast)
            self.ct.fix.statement_terminator = True
            terminator_tokens.append(self.ct)

        if not terminator_tokens:
            # We found nothing. This is actually a syntax error in
            # most cases.
            if allow_nothing:
                ending_token.fix.flag_continuations = True
                if self.cfg.active("end_of_statements"):
                    if semi:
                        ending_token.add_semicolon_after = True
                        if self.cfg.active("indentation"):
                            ending_token.fix.add_newline = True
                        self.mh.style_issue(ending_token.location,
                                            "end this with a ; and newline",
                                            "end_of_statements",
                                            True)
                    else:
                        fixed = False
                        if self.cfg.active("indentation"):
                            ending_token.fix.add_newline = True
                            fixed = True
                        self.mh.style_issue(ending_token.location,
                                            "end statement with a newline",
                                            "end_of_statements",
                                            fixed)
                return
            elif self.peek_eof():
                # EOF is also a valid (but rude) terminator
                return
            else:
                self.mh.error(self.nt.location,
                              "expected end of statement,"
                              " found %s instead" % self.nt.kind)
            raise ICE("logic error")
        assert len(terminator_tokens) >= 1

        if not self.cfg.active("end_of_statements"):
            return

        if semi:
            # Exactly two tokens are required and useful. The first
            # semicolon, and the first new_line.
            if terminator_tokens[0].kind == "SEMICOLON":
                pass

            elif terminator_tokens[0].kind == "COMMA":
                self.mh.style_issue(terminator_tokens[0].location,
                                    "end this with a semicolon"
                                    " instead of a comma",
                                    "end_of_statements",
                                    True)
                terminator_tokens[0].fix.change_to_semicolon = True

            else:
                assert terminator_tokens[0].kind == "NEWLINE"
                self.mh.style_issue(ending_token.location,
                                    "end statement with a semicolon",
                                    "end_of_statements",
                                    True)
                ending_token.fix.add_semicolon_after = True

            if first_newline is None:
                fixed = False
                if self.cfg.active("indentation"):
                    terminator_tokens[0].fix.add_newline = True
                    fixed = True
                self.mh.style_issue(terminator_tokens[0].location,
                                    "end statement with a newline",
                                    "end_of_statements",
                                    fixed)

        else:
            # Exactly one token is required and useful. The first new
            # line.
            if not self.cfg.active("indentation") and \
               terminator_tokens[0].kind == "COMMA":
                # The statement was ended with a comma. Ideally we
                # just have a newline, but since we're not fixing
                # indetation we can't fix it. Complain instead about
                # the comma
                self.mh.style_issue(ending_token.location,
                                    "end this with just a newline",
                                    "end_of_statements",
                                    False)
                if first_newline is None:
                    terminator_tokens[0].fix.change_to_semicolon = True

            elif terminator_tokens[0].kind != "NEWLINE":
                fixed = False
                if first_newline is None:
                    # We can only fix a missing newline if indentation
                    # fixing is active.
                    if self.cfg.active("indentation"):
                        terminator_tokens[0].fix.delete = True
                        ending_token.fix.add_newline = True
                        fixed = True
                else:
                    terminator_tokens[0].fix.delete = True
                    fixed = True
                self.mh.style_issue(terminator_tokens[0].location,
                                    "end this with just a newline",
                                    "end_of_statements",
                                    fixed)

        for terminator in terminator_tokens[1:]:
            if terminator.kind != "NEWLINE":
                self.mh.style_issue(terminator.location,  # Molten steel?
                                    "unnecessary statement terminator",
                                    "end_of_statements",
                                    True)
                terminator.fix.delete = True

    def parse_identifier(self, allow_void, allow_some_keywords=False):
        # identifier ::= <IDENTIFIER>
        #
        # void_or_identifier ::= identifier
        #                      | '~'
        #
        # If allow_some_keywords is true, we permit some keywords that
        # are normally reserved to be identifiers. This is to allow
        # class methods called 'end' or 'methods'.
        #
        # The exception to this is 'end', since that is generally
        # allowed since it makes parsing expressions using ranges much
        # easier.
        if self.peek("OPERATOR", "~") and allow_void:
            self.match("OPERATOR")
        elif self.peek("KEYWORD", "end"):
            self.match("KEYWORD", "end")
        elif allow_some_keywords and self.peek("KEYWORD", "import"):
            self.match("KEYWORD", "import")
        elif allow_some_keywords and self.peek("KEYWORD", "arguments"):
            self.match("KEYWORD", "arguments")
        else:
            self.match("IDENTIFIER")

        return Identifier(self.ct)

    def parse_name(self, allow_void):
        # superclass_ref ::= simple_name '@' function_reference
        #
        # simple_name ::= identifier
        #               | simple_name '.' identifier
        #
        # function_reference ::= simple_name
        #                      | simple_name '(' expression_list ')'
        #
        # name ::= superclass_ref
        #        | simple_name
        #        | name '.' identifier
        #        | name '.' '(' expression ')'
        #        | name '(' expression_list ')'
        #        | name '{' expression_list '}'
        #
        # Note that we can only resolve the ambiguity between
        # metaclass, simplename or any of the others late (but we can
        # always resolve it).
        #
        # expression_list ::= <>
        #                   | expression { ',' expression }

        # First we parse as much as possible as a simple name.
        rv = self.parse_simple_name(allow_void)

        # Then we can see if we have a superclass reference. What
        # follows are pretty different parse rules for the two cases.

        if self.peek("AT"):
            self.match("AT")
            t_at = self.ct
            at_prefix = rv
            at_suffix = self.parse_simple_name()
            if self.peek("BRA"):
                at_suffix = Reference(at_suffix)
                at_suffix.set_arguments(self.parse_argument_list(at_suffix))

            return Superclass_Reference(t_at, at_prefix, at_suffix)

        else:
            while (self.peek("SELECTION") or
                   self.peek("BRA") or
                   self.peek("C_BRA")):
                if self.peek("SELECTION"):
                    self.match("SELECTION")
                    tok = self.ct

                    if self.peek("BRA"):
                        self.match("BRA")
                        t_open = self.ct
                        dyn_field = self.parse_expression()
                        self.match("KET")
                        t_close = self.ct
                        rv = Dynamic_Selection(tok, rv, dyn_field)
                        t_open.set_ast(rv)
                        t_close.set_ast(rv)
                    else:
                        field = self.parse_identifier(allow_void=False)
                        rv = Selection(tok, rv, field)
                elif self.peek("BRA"):
                    rv = Reference(rv)
                    rv.set_arguments(self.parse_argument_list(rv))
                elif self.peek("C_BRA"):
                    rv = self.parse_cell_reference(rv)
                else:
                    raise ICE("impossible path (nt.kind = %s)" % self.nt.kind)

            return rv

    def parse_simple_name(self, allow_void=False, allow_some_keywords=False):
        # reference ::= identifier
        #             | reference '.' identifier

        rv = self.parse_identifier(allow_void=allow_void,
                                   allow_some_keywords=allow_some_keywords)

        # We need to lookahead 2 here to avoid parsing dynamic fields

        while self.peek("SELECTION") and not self.peek2("BRA"):
            if self.peek("SELECTION"):
                self.match("SELECTION")
                tok = self.ct
                field = self.parse_identifier(
                    allow_void=allow_void,
                    allow_some_keywords=allow_some_keywords)
                rv = Selection(tok, rv, field)
            else:
                raise ICE("impossible path (nt.kind = %s)" % self.nt.kind)

        return rv

    def parse_file(self):
        # This is the top-level parse function. First we need to
        # figure out exactly what kind of file we're dealing
        # with. This also hilariously depends on the file name.

        # File can start with pragmas or newlines. We need to process
        # them first until we arrive at the first interesting thing
        # that helps us decide.
        l_pragmas = []
        while self.peek("NEWLINE") or self.apeek("KEYWORD", "pragma"):
            if self.peek("NEWLINE"):
                self.skip()
            else:
                l_pragmas.append(self.parse_annotation_pragma())

        if self.peek("KEYWORD", "function"):
            l_functions, l_more_pragmas = self.parse_function_list()
            cunit = Function_File(os.path.basename(self.lexer.filename),
                                  os.path.dirname(
                                      pathutil.abspath(self.lexer.filename)),
                                  self.lexer.get_file_loc(),
                                  self.lexer.line_count(),
                                  l_functions,
                                  self.lexer.in_class_directory,
                                  l_pragmas + l_more_pragmas)
        elif self.peek("KEYWORD", "classdef"):
            cunit = self.parse_class_file(l_pragmas)
        elif self.language.script_global_functions:
            cunit = self.parse_octave_script_file(l_pragmas)
        else:
            cunit = self.parse_matlab_script_file(l_pragmas)

            # Special detection of Octave inline globals in MATLAB
            # mode (to improve the messages produced in #198).
            if not self.peek_eof() and cunit.l_functions:
                self.mh.error(cunit.l_functions[0].loc(),
                              "script-global functions are an Octave-specific"
                              " feature; move your functions to the end of"
                              " the script file or use an Octave language")

        if self.debug_tree:
            cunit.debug_parse_tree()

        self.match_eof()

        return cunit

    def parse_matlab_script_file(self, l_pragmas):
        # A MATLAB script file is a bunch of statements, followed by
        # zero or more private functions. This is distinct from Octave
        # script files, which allow you to sprinkle functions anywhere
        # in the top-level context.

        # At least in MATLAB 2017b script files cannot use the
        # non-ended chained functions.

        self.functions_require_end = True
        statements = []
        while not self.peek_eof():
            if self.peek("KEYWORD", "function"):
                break
            else:
                statements.append(self.parse_statement())

        l_functions, l_more_pragmas = self.parse_function_list()

        return Script_File(os.path.basename(self.lexer.filename),
                           os.path.dirname(
                               pathutil.abspath(self.lexer.filename)),
                           self.lexer.get_file_loc(),
                           self.lexer.line_count(),
                           Sequence_Of_Statements(statements),
                           l_functions,
                           l_pragmas + l_more_pragmas)

    def parse_octave_script_file(self, l_pragmas):
        # An Octave script file is very similar to a MATLAB one,
        # except they allow you to put functions everywhere (as long
        # as it's not the first thing, because then you have a
        # function file again). The semantics are weird as well,
        # because all functions encountered this way are made global,
        # as if they were in their own function files.

        self.functions_require_end = True
        statements = []
        l_functions = []
        l_more_pragmas = []

        if not self.peek_eof():
            # Script files _have_ to start with a statment
            statements.append(self.parse_statement())

        # Otherwise, in Octave, script files are a sequence of
        # statements intermixed with (global) function definitions
        while not self.peek_eof():
            if self.peek("KEYWORD", "function"):
                l_functions.append(self.parse_function_def(in_class = False))
            elif self.peek("KEYWORD", "pragma"):
                l_more_pragmas.append(self.parse_annotation_pragma())
            else:
                statements.append(self.parse_statement())

        return Script_File(os.path.basename(self.lexer.filename),
                           os.path.dirname(
                               pathutil.abspath(self.lexer.filename)),
                           self.lexer.get_file_loc(),
                           self.lexer.line_count(),
                           Sequence_Of_Statements(statements),
                           l_functions,
                           l_pragmas + l_more_pragmas)

    def parse_class_file(self, l_pragmas):
        self.functions_require_end = True

        n_classdef                  = self.parse_classdef()

        if self.language.allow_classdef_subfunctions:
            l_functions, l_more_pragmas = self.parse_function_list()
        else:
            l_functions    = []
            l_more_pragmas = []

        return Class_File(os.path.basename(self.lexer.filename),
                          os.path.dirname(
                              pathutil.abspath(self.lexer.filename)),
                          self.lexer.get_file_loc(),
                          self.lexer.line_count(),
                          n_classdef,
                          l_functions,
                          l_pragmas + l_more_pragmas)

    def parse_function_list(self):
        l_functions = []
        l_pragmas = []
        while self.peek("KEYWORD", "function") or \
              self.apeek("KEYWORD", "pragma"):
            if self.peek("KEYWORD", "function"):
                l_functions.append(self.parse_function_def(in_class = False))
            else:
                l_pragmas.append(self.parse_annotation_pragma())

        if not self.functions_require_end and l_functions:
            if len(l_functions) > 1:
                raise ICE("logic error")
            l_functions = self.reorder_as_function_list(l_functions[0])

        return l_functions, l_pragmas

    def reorder_as_function_list(self, n_fdef):
        # To deal with the special case where none of the functions
        # are terminated by end we need to flatten out the list we
        # have.
        assert isinstance(n_fdef, Function_Definition)
        functions = []

        while n_fdef:
            functions.append(n_fdef)
            if len(n_fdef.l_nested) == 1:
                n_fdef = n_fdef.l_nested.pop()
            elif len(n_fdef.l_nested) > 1:
                raise ICE("logic error")
            else:
                break

        return functions

    def parse_function_signature(self, in_class):
        assert isinstance(in_class, bool)

        rv = Function_Signature()

        # Parse returns. Either 'x' or a list '[x, y]'
        l_outputs = []
        if self.peek("A_BRA"):
            out_brackets = True
            self.match("A_BRA")
            self.ct.set_ast(rv)

            if not self.peek("A_KET"):
                while True:
                    l_outputs.append(self.parse_identifier(allow_void=True))
                    if self.peek("COMMA"):
                        self.match("COMMA")
                        self.ct.set_ast(rv)
                    else:
                        break

            self.match("A_KET")
            self.ct.set_ast(rv)

        else:
            out_brackets = False
            l_outputs.append(self.parse_simple_name())

        if self.peek("BRA") and len(l_outputs) == 1 and not out_brackets:
            # This is a function that doesn't return anything, so
            # function foo(...
            n_name = l_outputs[0]
            l_outputs = []

        elif self.peek("NEWLINE") and len(l_outputs) == 1 and not out_brackets:
            # As above, but without the brackets
            n_name = l_outputs[0]
            l_outputs = []

        else:
            # This is a normal function, so something like
            # function [a, b] = potato...
            # function a = potato...
            self.match("ASSIGNMENT")
            self.ct.set_ast(rv)
            n_name = self.parse_simple_name(allow_some_keywords=True)

        l_inputs = []
        if self.peek("BRA"):
            self.match("BRA")
            self.ct.set_ast(rv)

            if not self.peek("KET"):
                while True:
                    l_inputs.append(self.parse_identifier(allow_void=True))
                    if self.peek("COMMA"):
                        self.match("COMMA")
                        self.ct.set_ast(rv)
                    else:
                        break

            self.match("KET")
            self.ct.set_ast(rv)

        if isinstance(n_name, Selection) and not in_class:
            self.mh.error(n_name.loc(),
                          "dotted name is not permitted outside class")

        rv.set_name(n_name)
        rv.set_inputs(l_inputs)
        rv.set_outputs(l_outputs)
        self.match_eos(rv)
        return rv

    def parse_function_def(self, in_class):
        assert isinstance(in_class, bool)

        self.match("KEYWORD", "function")
        self.push_context("function")
        t_fun = self.ct

        n_sig = self.parse_function_signature(in_class)

        l_body = []
        l_nested = []
        l_argval = []

        # First, deal with any argument validation blocks
        while self.peek("KEYWORD", "arguments"):
            l_argval.append(self.parse_validation_block())

        # Then, process the rest of the function
        while not self.peek("KEYWORD", "end") and not self.peek_eof():
            item = self.parse_statement()
            if isinstance(item, Function_Definition):
                l_nested.append(item)
            else:
                l_body.append(item)

        rv = Function_Definition(t_fun, n_sig,
                                 l_argval,
                                 Sequence_Of_Statements(l_body),
                                 l_nested)

        if self.peek_eof() and self.functions_require_end:
            self.mh.error(t_fun.location,
                          "this function must be terminated with end")
        elif self.peek_eof():
            # TODO: style issue
            pass
        else:
            self.functions_require_end = True
            self.match("KEYWORD", "end")
            rv.set_end(self.ct)
            self.match_eos(rv)

        self.pop_context()

        return rv

    def parse_name_value_pair_list(self, n_ast):
        assert isinstance(n_ast, Node)

        properties = []

        if self.peek("BRA"):
            self.match("BRA")
            self.ct.set_ast(n_ast)
            while True:
                n_pair = Name_Value_Pair(
                    self.parse_identifier(allow_void=False))

                if self.peek("ASSIGNMENT"):
                    self.match("ASSIGNMENT")
                    t_eq = self.ct
                    n_value = self.parse_expression()
                    n_pair.set_value(t_eq, n_value)

                properties.append(n_pair)

                if self.peek("COMMA"):
                    self.match("COMMA")
                    self.ct.set_ast(n_ast)
                else:
                    break
            self.match("KET")
            self.ct.set_ast(n_ast)

        return properties

    def parse_validation_block(self):
        # See
        # https://uk.mathworks.com/help/matlab/matlab_oop/validate-property-values.html
        # https://uk.mathworks.com/help/matlab/matlab_oop/property-validator-functions.html
        # https://www.mathworks.com/help/matlab/matlab_prog/function-argument-validation-1.html

        if self.peek("KEYWORD", "arguments"):
            self.match("KEYWORD", "arguments")
        else:
            self.match("KEYWORD", "properties")
        t_kw = self.ct

        rv = Special_Block(t_kw)
        rv.set_attributes(self.parse_name_value_pair_list(rv))
        self.match_eos(rv)

        while not self.peek("KEYWORD", "end"):
            # First the name we refer to
            if t_kw.value == "arguments":
                n_name = self.parse_simple_name(allow_void=True)
            else:
                n_name = self.parse_identifier(allow_void=False)

            # We can have a delegation, or (more likely) an ordinary
            # constraint.
            if t_kw.value == "arguments" and self.peek("NVP_DELEGATE"):
                self.match("NVP_DELEGATE")

                delegation = Argument_Validation_Delegation(self.ct)
                delegation.set_name(n_name)
                delegation.set_class_name(self.parse_simple_name())

                self.match_eos(delegation)

                rv.add_delegation(delegation)

            else:
                cons = Entity_Constraints()
                cons.set_name(n_name)

                # All other validation options are optional. Historically
                # this is likely because the properties block used to be
                # just a list of class properties, and then it evolved
                # (and eventually got re-used in function argument
                # validation blocks).

                # Dimension validation
                val_dim = []
                if self.peek("BRA"):
                    self.match("BRA")
                    self.ct.set_ast(cons)

                    while True:
                        if self.peek("NUMBER"):
                            self.match("NUMBER")
                            val_dim.append(self.ct)
                        elif self.peek("COLON"):
                            self.match("COLON")
                            val_dim.append(self.ct)
                        else:
                            self.mh.error(self.nt.location,
                                          "dimension validation may contain"
                                          " only integral numbers or :")

                        if self.peek("COMMA"):
                            self.match("COMMA")
                            self.ct.set_ast(cons)
                        else:
                            break

                    self.match("KET")
                    self.ct.set_ast(cons)

                if len(val_dim) == 1:
                    self.mh.error(self.ct.location,
                                  "in MATLAB dimension constraints must"
                                  " contain at least two dimensions",
                                  fatal=False)
                    # I speculate that the underlying reason is that the
                    # MATLAB parser otherwise sees this as a function or
                    # index? Reasons very unclear, this is just a wild
                    # guess.
                elif len(val_dim) >= 2:
                    cons.set_dimension_constraints(val_dim)

                # Class validation
                if self.peek("IDENTIFIER"):
                    cons.set_class_constraint(self.parse_simple_name())

                # Function validation
                if self.peek("C_BRA"):
                    self.match("C_BRA")
                    self.ct.set_ast(cons)

                    while not self.peek("C_KET"):
                        cons.add_functional_constraint(
                            self.parse_name(allow_void=False))
                        if self.peek("COMMA"):
                            self.match("COMMA")
                            self.ct.set_ast(cons)
                        else:
                            break

                    self.match("C_KET")
                    self.ct.set_ast(cons)

                # Default value
                if self.peek("ASSIGNMENT"):
                    self.match("ASSIGNMENT")
                    self.ct.set_ast(cons)
                    cons.set_default_value(self.parse_expression())

                self.match_eos(cons)

                rv.add_constraint(cons)

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)

        return rv

    def parse_class_methods(self):
        # Using:
        # https://www.mathworks.com/help/matlab/matlab_oop/specifying-methods-and-functions.html
        # https://www.mathworks.com/help/matlab/matlab_oop/method-attributes.html
        # https://www.mathworks.com/help/matlab/matlab_oop/methods-in-separate-files.html

        self.match("KEYWORD", "methods")
        t_kw = self.ct

        rv = Special_Block(t_kw)
        rv.set_attributes(self.parse_name_value_pair_list(rv))
        self.match_eos(rv)

        while not self.peek("KEYWORD", "end"):
            if self.peek("KEYWORD", "function"):
                rv.add_method(self.parse_function_def(in_class = True))
            else:
                rv.add_method(self.parse_function_signature(in_class = True))

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)

        return rv

    def parse_enumeration(self):
        # Using:
        # https://uk.mathworks.com/help/matlab/matlab_oop/enumerations.html

        self.match("KEYWORD", "enumeration")
        t_kw = self.ct

        rv = Special_Block(t_kw)
        self.match_eos(rv, allow_nothing=True)

        while not self.peek("KEYWORD", "end"):
            enum = Class_Enumeration(self.parse_identifier(allow_void=False))

            if self.peek("BRA"):
                self.match("BRA")
                self.ct.set_ast(enum)

                while not self.peek("KET"):
                    enum.add_argument(self.parse_expression())
                    if self.peek("COMMA"):
                        self.match("COMMA")
                        self.ct.set_ast(enum)
                    else:
                        break

                self.match("KET")
                self.ct.set_ast(enum)

            rv.add_enumeration(enum)

            self.match_eos(rv)

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)

        return rv

    def parse_class_events(self):
        # Using the syntax described in
        # https://www.mathworks.com/help/matlab/matlab_oop/events-and-listeners.html
        self.match("KEYWORD", "events")
        t_kw = self.ct

        rv = Special_Block(t_kw)
        self.match_eos(rv)

        while not self.peek("KEYWORD", "end"):
            rv.add_event(self.parse_identifier(allow_void=False))
            self.match_eos(rv)

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)

        return rv

    def parse_classdef(self):
        # Using the syntax described in
        # https://uk.mathworks.com/help/matlab/matlab_oop/user-defined-classes.html
        # https://uk.mathworks.com/help/matlab/matlab_oop/class-components.html

        self.match("KEYWORD", "classdef")
        self.push_context("classdef")
        rv = Class_Definition(self.ct)

        # Class attributes. Ignored for now.
        rv.set_attributes(self.parse_name_value_pair_list(rv))

        # Class name
        rv.set_name(self.parse_identifier(allow_void=False))

        # Inheritance
        l_super = []
        if self.peek("OPERATOR", "<"):
            self.match("OPERATOR", "<")
            self.ct.set_ast(rv)

            while True:
                sc_name = self.parse_simple_name()
                l_super.append(sc_name)

                if self.peek("OPERATOR", "&"):
                    self.match("OPERATOR", "&")
                    self.ct.set_ast(rv)
                else:
                    break
        rv.set_super_classes(l_super)

        self.match_eos(rv)

        while True:
            if self.peek("KEYWORD", "properties"):
                rv.add_block(self.mh, self.parse_validation_block())
            elif self.peek("KEYWORD", "methods"):
                rv.add_block(self.mh, self.parse_class_methods())
            elif self.peek("KEYWORD", "events"):
                rv.add_block(self.mh, self.parse_class_events())
            elif self.peek("KEYWORD", "enumeration"):
                rv.add_block(self.mh, self.parse_enumeration())
            elif self.peek("KEYWORD", "end"):
                break
            else:
                self.mh.error(self.nt.location,
                              "expected properties|methods|events|enumeration"
                              " inside classdef")

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)
        self.pop_context()

        return rv

    def parse_delimited_input(self):
        statements = []

        while True:
            if self.peek("KEYWORD") and self.nt.value in ("end",
                                                          "catch",
                                                          "case",
                                                          "otherwise",
                                                          "else",
                                                          "elseif"):
                break
            statements.append(self.parse_statement())

        return Sequence_Of_Statements(statements)

    def parse_annotation_static_string_expression(self):
        self.amatch("STRING")
        rv = String_Literal(self.ct)

        while self.apeek("OPERATOR", "+"):
            self.amatch("OPERATOR", "+")
            t_op = self.ct

            self.amatch("STRING")
            rv = Binary_Operation(1, t_op, rv, String_Literal(self.ct))

        return rv

    def parse_annotation_pragma(self):
        punctuation = []
        n_reason    = None

        self.amatch("KEYWORD", "pragma")
        t_pragma = self.ct

        self.amatch("IDENTIFIER")
        t_pragma_kind = self.ct
        if t_pragma_kind.value not in ("Justify", "Tag", "No_Tracing"):
            self.mh.error(t_pragma_kind.location,
                          "unknown miss_hit pragma '%s'" %
                          t_pragma_kind.value)

        if t_pragma_kind.value in ("Justify", "Tag"):
            # Pragmas with parameters
            self.amatch("BRA")
            punctuation.append(self.ct)

            if t_pragma_kind.value == "Justify":
                self.amatch("IDENTIFIER")
                t_tool = self.ct

                if t_tool.value != "metric":
                    self.mh.error(t_tool.location,
                                  "unknown miss_hit tool '%s'" % t_tool.value)

                self.amatch("COMMA")
                punctuation.append(self.ct)

                self.amatch("STRING")
                t_param = self.ct

                if t_param.value not in METRICS:
                    self.mh.warning(t_param.location,
                                    "unknown metric '%s'" % t_param.value)

                self.amatch("COMMA")
                punctuation.append(self.ct)

                n_reason = self.parse_annotation_static_string_expression()

            elif t_pragma_kind.value == "Tag":
                l_tags = []
                while not self.apeek("KET"):
                    self.amatch("STRING")
                    l_tags.append(self.ct)
                    if self.apeek("COMMA"):
                        self.amatch("COMMA")
                        punctuation.append(self.ct)

            else:
                raise ICE("logic error")

            self.amatch("KET")
            punctuation.append(self.ct)

        self.amatch("SEMICOLON")
        punctuation.append(self.ct)

        if t_pragma_kind.value == "Justify":
            rv = Metric_Justification_Pragma(
                t_pragma, t_pragma_kind,
                t_tool, t_param, n_reason,
                self.cfg.style_config["regex_tickets"],
                self.cfg.ignore_pragmas_with_tickets)
        elif t_pragma_kind.value == "Tag":
            rv = Tag_Pragma(t_pragma, t_pragma_kind, l_tags)
        elif t_pragma_kind.value == "No_Tracing":
            rv = No_Tracing_Pragma(t_pragma, t_pragma_kind)
        else:
            raise ICE("logic error")

        for token in punctuation:
            token.set_ast(rv)

        return rv

    def parse_annotation_statement(self):
        if self.apeek("KEYWORD", "pragma"):
            return self.parse_annotation_pragma()
        else:
            self.mh.error(self.nt.location,
                          "expected valid miss_hit annotation statement")

    def parse_statement(self):
        if self.in_shortcircuit_context:
            raise ICE("failed to unset sc context")

        while self.peek("NEWLINE"):
            self.skip()

        if self.peek("KEYWORD"):
            if self.nt.value == "for":
                return self.parse_for_statement()
            elif self.nt.value == "if":
                return self.parse_if_statement()
            elif self.nt.value == "global":
                return self.parse_global_statement()
            elif self.nt.value == "while":
                return self.parse_while_statement()
            elif self.nt.value == "return":
                return self.parse_return_statement()
            elif self.nt.value == "switch":
                return self.parse_switch_statement()
            elif self.nt.value == "break":
                return self.parse_break_statement()
            elif self.nt.value == "continue":
                return self.parse_continue_statement()
            elif self.nt.value == "import":
                return self.parse_import_statement()
            elif self.nt.value == "try":
                return self.parse_try_statement()
            elif self.nt.value == "persistent":
                return self.parse_persistent_statement()
            elif self.nt.value == "parfor":
                return self.parse_parfor_statement()
            elif self.nt.value == "spmd":
                return self.parse_spmd_statement()
            elif self.nt.value == "function" and self.in_context("function"):
                if self.context[-1] != "function":
                    self.mh.error(self.nt.location,
                                  "nested function cannot appear inside"
                                  " %s" % self.context[-1])
                return self.parse_function_def(in_class = False)
            else:
                self.mh.error(self.nt.location,
                              "expected valid statement,"
                              " found keyword '%s' instead" % self.nt.value)
        elif self.peek_annotation():
            return self.parse_annotation_statement()
        elif self.peek("BANG"):
            self.match("BANG")
            t_bang = self.ct
            self.match("NEWLINE")
            return Naked_Expression_Statement(
                Function_Call(Identifier(t_bang),
                              [Char_Array_Literal(t_bang)],
                              "escape"))
        elif self.peek("A_BRA"):
            return self.parse_list_assignment()
        else:
            # This can be one of three things
            # other_stmt ::= reference "=" expr # simple assignment
            #              | reference CARRAY+  # command form
            #              | expression         # naked expression, could be
            #                                   # a call
            rv = self.parse_expression()

            if self.peek("ASSIGNMENT"):
                self.match("ASSIGNMENT")
                t_eq = self.ct
                if not isinstance(rv, Name):
                    self.mh.error(t_eq.location,
                                  "left-hand side of assignment must be a"
                                  " Name, found %s instead" %
                                  rv.__class__.__name__)
                rhs = self.parse_nested_expression()
                if self.cfg.active("builtin_shadow"):
                    rv.sty_check_builtin_shadow(self.mh, self.cfg)
                rv = Simple_Assignment_Statement(t_eq, rv, rhs)

            elif self.peek("CARRAY"):
                # Sanity check that the function is a simple name
                if not isinstance(rv, (Identifier, Selection)):
                    self.mh.error(self.ct.location,
                                  "command form requires a simple (dotted)"
                                  " identifier; found %s instead" %
                                  rv.__class__.__name__)

                arg_list = []
                while self.peek("CARRAY"):
                    self.match("CARRAY")
                    arg_list.append(Char_Array_Literal(self.ct))
                rv = Function_Call(rv, arg_list, "command")
                rv = Naked_Expression_Statement(rv)

            else:
                rv = Naked_Expression_Statement(rv)

            self.match_eos(rv, ";")

            return rv

    def parse_list_assignment(self):
        # Assignment
        #   s_assignee_matrix "=" expr           '[<ref>] = <expr>'
        #   m_assignee_matrix "=" reference      '[<ref>, <ref>] = <ref>'
        #
        # The list cannot be empty, and there can be an optional
        # trailing comma. This is not the same as a function return
        # list, since there cannot be a trailing comma there.

        rv = Compound_Assignment_Statement()

        lhs = []
        require_comma = False

        self.match("A_BRA")
        self.ct.set_ast(rv)
        if self.peek("COMMA"):
            self.match("COMMA")
            self.ct.set_ast(rv)
        while True:
            # There is a special case we need to take care of with
            # ~. There is a MATLAB bug/weirdness with [~ x], which is
            # parsed like [~x], but [x y] is OK for some reason. See
            # issue #70. Hence we enforce commas after any ~.
            if self.peek("OPERATOR", "~"):
                require_comma = True
            target = self.parse_name(allow_void=True)
            if self.cfg.active("builtin_shadow"):
                target.sty_check_builtin_shadow(self.mh, self.cfg)
            lhs.append(target)
            if (self.peek("COMMA") or require_comma) and \
               not self.peek("A_KET"):
                self.match("COMMA")
                self.ct.set_ast(rv)
                require_comma = False
            if self.peek("A_KET"):
                break
        self.match("A_KET")
        self.ct.set_ast(rv)
        rv.set_targets(lhs)

        self.match("ASSIGNMENT")
        rv.set_token_eq(self.ct)

        if len(lhs) == 1:
            # We've got something like
            #    [x] = <expr>
            rhs = self.parse_expression()
        else:
            # We've got something like
            #    [x, y] = fun(...)
            #
            # I believe that this can't be an expression, it basically
            # has to be a function call. Needs to be checked.
            rhs = self.parse_expression()
        rv.set_expression(rhs)

        self.match_eos(rv, ";")
        return rv

    def parse_nested_expression(self):
        return self.parse_precedence_12()

    def parse_expression(self):
        n_expr = self.parse_nested_expression()
        self.check_redundant_brackets(n_expr)
        return n_expr

    # 1. Parentheses ()
    def parse_precedence_1(self):
        if self.peek("NUMBER"):
            self.match("NUMBER")
            return Number_Literal(self.ct)

        elif self.peek("CARRAY"):
            self.match("CARRAY")
            return Char_Array_Literal(self.ct)

        elif self.peek("STRING"):
            self.match("STRING")
            return String_Literal(self.ct)

        elif self.peek("BRA"):
            self.match("BRA")
            t_open = self.ct
            expr = self.parse_nested_expression()
            self.match("KET")
            t_close = self.ct
            self.set_expression_brackets(expr, t_open, t_close)
            return expr

        elif self.peek("M_BRA"):
            with self.sc_context(False):
                return self.parse_matrix()

        elif self.peek("C_BRA"):
            with self.sc_context(False):
                return self.parse_cell()

        elif self.peek("COLON"):
            self.match("COLON")
            return Reshape(self.ct)

        elif self.peek("AT"):
            return self.parse_function_handle()

        elif self.peek("METACLASS"):
            self.match("METACLASS")
            tok = self.ct
            return Metaclass(tok, self.parse_simple_name())

        else:
            with self.sc_context(False):
                return self.parse_name(allow_void=False)

    # 2. Transpose (.'), power (.^), complex conjugate transpose ('),
    #    matrix power (^)
    #
    # 3. Power with unary minus (.^-), unary plus (.^+), or logical
    #    negation (.^~) as well as matrix power with unary minus (^-), unary
    #    plus (^+), or logical negation (^~).
    #
    #    Note: Although most operators work from left to right, the
    #    operators (^-), (.^-), (^+), (.^+), (^~), and (.^~) work from
    #    second from the right to left. It is recommended that you use
    #    parentheses to explicitly specify the intended precedence of
    #    statements containing these operator combinations.
    def parse_precedence_2(self):
        # In Octave chaining ^ is left associative, i.e. 2 ^ 3 ^ 2 ==
        # (2 ^ 3) ^ 2 == 64.
        #
        # TODO: Is this also true for MATLAB?
        rv = self.parse_precedence_1()

        while self.peek("OPERATOR") and self.nt.value in ("^", ".^",
                                                          "'", ".'"):
            self.match("OPERATOR")
            t_op = self.ct
            if t_op.value in ("^", ".^"):
                unary_chain = []
                while self.peek("OPERATOR") and \
                      self.nt.value in ("-", "+", "~"):
                    self.match("OPERATOR")
                    unary_chain.append(self.ct)
                rhs = self.parse_precedence_1()
                while unary_chain:
                    rhs = Unary_Operation(3, unary_chain.pop(), rhs)
                rv = Binary_Operation(2, t_op, rv, rhs)
            else:
                rv = Unary_Operation(2, t_op, rv)

        return rv

    def parse_precedence_3(self):
        # This is dealt with as a special case in (2).
        return self.parse_precedence_2()

    # 4. Unary plus (+), unary minus (-), logical negation (~)
    #
    # In Octave ! is also the the logical negation operator.
    def parse_precedence_4(self):
        if self.peek("OPERATOR") and self.nt.value in ("+", "-", "~", "!"):
            self.match("OPERATOR")
            t_op = self.ct
            rhs = self.parse_precedence_4()
            return Unary_Operation(4, t_op, rhs)
        else:
            return self.parse_precedence_3()

    # 5. Multiplication (.*), right division (./), left division (.\),
    #    matrix multiplication (*), matrix right division (/), matrix left
    #    division (\)
    def parse_precedence_5(self):
        rv = self.parse_precedence_4()

        while self.peek("OPERATOR") and self.nt.value in ("*", ".*",
                                                          "/", "./",
                                                          "\\", ".\\"):
            self.match("OPERATOR")
            t_op = self.ct
            rhs = self.parse_precedence_4()
            rv = Binary_Operation(5, t_op, rv, rhs)

        return rv

    # 6. Addition (+), subtraction (-)
    def parse_precedence_6(self):
        rv = self.parse_precedence_5()

        while self.peek("OPERATOR") and self.nt.value in ("+", "-"):
            self.match("OPERATOR")
            t_op = self.ct
            rhs = self.parse_precedence_5()
            rv = Binary_Operation(6, t_op, rv, rhs)

        return rv

    # 7. Colon operator (:)
    def parse_range_expression(self):
        t_first_colon = None
        t_second_colon = None
        points = []
        points.append(self.parse_precedence_6())
        if self.peek("COLON"):
            self.match("COLON")
            t_first_colon = self.ct
            points.append(self.parse_precedence_6())
        if self.peek("COLON"):
            self.match("COLON")
            t_second_colon = self.ct
            points.append(self.parse_precedence_6())
        assert 1 <= len(points) <= 3

        if len(points) == 1:
            return points[0]
        elif len(points) == 2:
            return Range_Expression(points[0], t_first_colon, points[1])
        else:
            return Range_Expression(points[0], t_first_colon, points[2],
                                    t_second_colon, points[1])

    # 8. Less than (<), less than or equal to (<=), greater than (>),
    #    greater than or equal to (>=), equal to (==), not equal to (~=)
    #
    # In Octave we also get != as not equal
    def parse_precedence_8(self):
        rv = self.parse_range_expression()

        chain_length = 1
        while self.peek("OPERATOR") and self.nt.value in ("<", "<=",
                                                          ">", ">=",
                                                          "==", "~=",
                                                          "!="):
            chain_length += 1
            self.match("OPERATOR")
            t_op = self.ct
            rhs = self.parse_range_expression()
            rv = Binary_Operation(8, t_op, rv, rhs)

            if chain_length > 2:
                self.mh.check(t_op.location,
                              "chained relation does not work the"
                              " way you think it does",
                              "chained_relation",
                              "high")

        return rv

    # 9. Element-wise AND (&)
    def parse_precedence_9(self):
        rv = self.parse_precedence_8()

        while self.peek("OPERATOR", "&"):
            self.match("OPERATOR", "&")
            t_op = self.ct
            rhs = self.parse_precedence_8()
            rv = Binary_Logical_Operation(9,
                                          t_op, self.in_shortcircuit_context,
                                          rv, rhs)

        return rv

    # 10. Element-wise OR (|)
    def parse_precedence_10(self):
        rv = self.parse_precedence_9()

        while self.peek("OPERATOR", "|"):
            self.match("OPERATOR", "|")
            t_op = self.ct
            rhs = self.parse_precedence_9()
            rv = Binary_Logical_Operation(10,
                                          t_op, self.in_shortcircuit_context,
                                          rv, rhs)

        return rv

    # 11. Short-circuit AND (&&)
    def parse_precedence_11(self):
        rv = self.parse_precedence_10()

        while self.peek("OPERATOR", "&&"):
            self.match("OPERATOR", "&&")
            t_op = self.ct
            rhs = self.parse_precedence_10()
            rv = Binary_Logical_Operation(11, t_op, True, rv, rhs)

        return rv

    # 12. Short-circuit OR (||)
    def parse_precedence_12(self):
        rv = self.parse_precedence_11()

        while self.peek("OPERATOR", "||"):
            self.match("OPERATOR", "||")
            t_op = self.ct
            rhs = self.parse_precedence_11()
            rv = Binary_Logical_Operation(12, t_op, True, rv, rhs)

        return rv

    def parse_row_list(self):
        # Parse the contents of a matrix or cell expression.

        # matrix_expression ::= '[' row_list ']'
        # cell_expression ::= '{' row_list '}'
        #
        # row_list ::= row { row_separation {row_separation} row_list }
        #
        # row_separation ::=  ';' | '\n'

        rv = Row_List()

        # Loop until we hit the end
        while not (self.peek_eof() or
                   self.peek("C_KET") or
                   self.peek("M_KET")):
            # row
            current_row = self.parse_row()
            rv.add_item(current_row)

            # row_separation {row_separation}

            # We keep a track of all semicolons so we can remove
            # some/all of them as spurious. We should also do that to
            # new-lines, but right now adding/removing lines is really
            # iffy.
            separation_tokens = []
            first_semi = None
            first_nl = None
            if self.peek("NEWLINE") or self.peek("SEMICOLON"):
                while self.peek("NEWLINE") or self.peek("SEMICOLON"):
                    if self.peek("NEWLINE") and first_nl is None:
                        first_nl = self.nt
                    elif self.peek("SEMICOLON"):
                        if first_semi is None:
                            first_semi = self.nt
                        separation_tokens.append(self.nt)
                    self.skip()
                    self.ct.set_ast(rv)
            else:
                # Any other token means we did not find a row
                # separation, so the row-list is over
                break

            # there are two canonical forms to separate rows:
            #   * a single newline
            #   * a single semicolon
            # we prefer the first if there is at least one new-line

            if first_nl is not None or \
               current_row.is_empty() or \
               self.peek_eof() or \
               self.peek("C_KET") or \
               self.peek("M_KET"):
                # This is the case where we are already at the end of
                # the matrix/cell, or we had an entirely empty row, or
                # we have a newline: any semicolon is spurious
                for token in separation_tokens:
                    token.fix.spurious = True
            else:
                # All tokens except the first semicolon are spurious
                for token in separation_tokens:
                    if token is first_semi:
                        pass
                    else:
                        token.fix.spurious = True

        return rv

    def parse_row(self):
        # At most one trailing and leading comma is allowed. Otherwise
        # this is just a comma-separated list.
        #
        # col_separation ::= ','
        #
        # expression_list ::= expression { col_separation expression }
        #
        # row ::= [col_separation] [expression_list [col_separation]]
        #
        # So this means all of the following are valid row examples:
        #   *     (the empty row)
        #   * ,   (a degenerate case)
        #   * ,1  (leading comma)
        #   * 1,  (trailing comma)
        #   * ,1, (leading and trailing comma)
        #   * 1,2 (normal case)
        #
        # But the following are not valid:
        #   * ,,
        #   * ,,1
        #   * 1,,

        rv = Row()

        # Detect the entirely blank row
        if self.peek("NEWLINE") or \
           self.peek("SEMICOLON") or \
           self.peek("C_KET") or \
           self.peek("M_KET"):
            return rv

        if self.peek("COMMA"):
            self.match("COMMA")
            self.ct.set_ast(rv)
            self.ct.fix.spurious = True

            # We'll parse an expression, unless it looks like we hit
            # the end of the matrix or cell, or the end of a row
            if self.peek("NEWLINE") or \
               self.peek("SEMICOLON") or \
               self.peek("COMMA") or \
               self.peek("C_KET") or \
               self.peek("M_KET"):
                return rv

        # Parse the expression list itself
        while True:
            rv.add_item(self.parse_nested_expression())

            if self.peek("COMMA"):
                self.match("COMMA")
                self.ct.set_ast(rv)

            # Detect trailing comma
            if self.peek("NEWLINE") or \
               self.peek("SEMICOLON") or \
               self.peek("C_KET") or \
               self.peek("M_KET"):
                self.ct.fix.spurious = True
                break

        return rv

    def parse_matrix(self):
        self.match("M_BRA")
        rv = Matrix_Expression(self.ct)
        rv.set_content(self.parse_row_list())
        self.match("M_KET")
        rv.set_closing_bracket(self.ct)
        return rv

    def parse_cell(self):
        self.match("C_BRA")
        rv = Cell_Expression(self.ct)
        rv.set_content(self.parse_row_list())
        self.match("C_KET")
        rv.set_closing_bracket(self.ct)
        return rv

    def parse_function_handle(self):
        self.match("AT")
        t_at = self.ct

        if self.peek("BRA"):
            rv = Lambda_Function(t_at)

            self.match("BRA")
            self.ct.set_ast(rv)

            while not self.peek("KET"):
                rv.add_parameter(self.parse_identifier(allow_void=True))
                if self.peek("COMMA"):
                    self.match("COMMA")
                    self.ct.set_ast(rv)
                else:
                    break

            self.match("KET")
            self.ct.set_ast(rv)

            rv.set_body(self.parse_nested_expression())
            return rv

        else:
            name = self.parse_simple_name()
            return Function_Pointer(t_at, name)

    def parse_argument_list(self, n_ast):
        assert isinstance(n_ast, Node)
        # arglist ::= '(' ')'
        #           | '(' expression { ',' expression } '}'
        #
        # Note: This list can be empty
        args = []
        self.match("BRA")
        self.ct.set_ast(n_ast)
        if self.peek("KET"):
            self.match("KET")
            self.ct.set_ast(n_ast)
            return args

        while True:
            if self.language.string_argument_pairs and \
               self.peek("IDENTIFIER") and self.peek2("ASSIGNMENT"):
                # Special case for the 2021a MATLAB mis-feature of
                # name-value pairs in function calls
                nv_pair = Name_Value_Pair(
                    self.parse_identifier(allow_void=False))
                self.match("ASSIGNMENT")
                t_eq = self.ct
                # We immeditely issue a Lint message about the
                # confusing semantics of = here.
                self.mh.check(self.ct.location,
                              "name-value pairs have extremely confusing"
                              " semantics and should be avoided, use two"
                              " arguments instead",
                              "name_value_pairs",
                              severity="low")

                nv_pair.set_value(t_eq, self.parse_expression())
                args.append(nv_pair)

            else:
                args.append(self.parse_expression())

            if self.peek("COMMA"):
                self.match("COMMA")
                self.ct.set_ast(n_ast)
            elif self.peek("KET"):
                break

        # At this point we need to make sure that any name-value pairs
        # provided are at the _end_ of the argument list.
        only_nv_pairs_allowed_now = False
        for n_arg in args:
            if isinstance(n_arg, Name_Value_Pair):
                only_nv_pairs_allowed_now = True
            elif only_nv_pairs_allowed_now:
                self.mh.error(n_arg.loc(),
                              "positional argument is not permitted after"
                              " a name-value pair")

        self.match("KET")
        self.ct.set_ast(n_ast)
        return args

    def parse_cell_reference(self, n_name):
        # cell_arglist ::= <<name>> '{' expression { ',' expression } '}'
        #
        # Note: cannot be empty
        rv = Cell_Reference(n_name)

        self.match("C_BRA")
        self.ct.set_ast(rv)

        while True:
            rv.add_argument(self.parse_expression())
            if self.peek("COMMA"):
                self.match("COMMA")
                self.ct.set_ast(rv)
            elif self.peek("C_KET"):
                break

        self.match("C_KET")
        self.ct.set_ast(rv)

        return rv

    def parse_if_statement(self):
        actions = []

        self.match("KEYWORD", "if")
        self.push_context("if")
        n_action = Action(self.ct)
        with self.sc_context(True):
            n_action.set_expression(self.parse_expression())
        self.match_eos(n_action, allow_nothing=True)
        n_action.set_body(self.parse_delimited_input())
        actions.append(n_action)

        while self.peek("KEYWORD", "elseif"):
            self.match("KEYWORD", "elseif")
            n_action = Action(self.ct)
            with self.sc_context(True):
                n_action.set_expression(self.parse_expression())
            self.match_eos(n_action, allow_nothing=True)
            n_action.set_body(self.parse_delimited_input())
            actions.append(n_action)

        if self.peek("KEYWORD", "else"):
            self.match("KEYWORD", "else")
            n_action = Action(self.ct)
            self.match_eos(n_action, allow_nothing=True)
            n_action.set_body(self.parse_delimited_input())
            actions.append(n_action)

        self.match("KEYWORD", "end")
        rv = If_Statement(actions)
        self.ct.set_ast(rv)
        self.match_eos(rv)
        self.pop_context()

        return rv

    def parse_return_statement(self):
        self.match("KEYWORD", "return")
        rv = Return_Statement(self.ct)
        self.match_eos(rv)

        return rv

    def parse_break_statement(self):
        self.match("KEYWORD", "break")
        rv = Break_Statement(self.ct)

        if not self.in_context("loop"):
            self.mh.error(self.ct.location,
                          "break must appear inside loop",
                          fatal = False)

        self.match_eos(rv)

        return rv

    def parse_continue_statement(self):
        self.match("KEYWORD", "continue")
        rv = Continue_Statement(self.ct)

        if not self.in_context("loop"):
            self.mh.error(self.ct.location,
                          "continue must appear inside loop",
                          fatal = False)

        self.match_eos(rv)

        return rv

    def parse_for_assignment(self,
                             n_ast,
                             allow_brackets=False,
                             allow_recursion=False):
        assert isinstance(n_ast, Node)
        # Apparently it's OK to wrap the loop in arbitrarily nested
        # brackets, e.g. for ((i = 1:2))
        #
        # TODO: Flag bad style
        #
        # TODO: In octave we can recurse for the normal for, in MATLAB
        # we cannot. For parfor we can never recurse.
        if self.peek("BRA") and allow_brackets:
            self.match("BRA")
            self.ct.set_ast(n_ast)
            n_ident, n_expr = self.parse_for_assignment(n_ast,
                                                        allow_recursion,
                                                        allow_recursion)
            self.match("KET")
            self.ct.set_ast(n_ast)
        else:
            n_ident = self.parse_identifier(allow_void=False)
            self.match("ASSIGNMENT")
            self.ct.set_ast(n_ast)
            n_expr = self.parse_expression()

        return n_ident, n_expr

    def parse_for_statement(self):
        self.match("KEYWORD", "for")
        self.push_context("loop")
        rv = General_For_Statement(self.ct)

        n_ident, n_expr = self.parse_for_assignment(rv, allow_brackets=True)
        rv.set_ident(n_ident)
        rv.set_expression(n_expr)
        self.match_eos(rv, allow_nothing=True)

        rv.set_body(self.parse_delimited_input())

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)
        self.pop_context()

        return rv

    def parse_parfor_statement(self):
        self.match("KEYWORD", "parfor")
        self.push_context("loop")
        rv = Parallel_For_Statement(self.ct)

        if self.peek("BRA"):
            # parfor (var = first:last, max_workers)
            self.match("BRA")
            self.ct.set_ast(rv)

            n_ident, n_expr = self.parse_for_assignment(rv)
            if self.peek("COMMA"):
                self.match("COMMA")
                self.ct.set_ast(rv)
                rv.set_workers(self.parse_expression())

            self.match("KET")
            self.ct.set_ast(rv)

        else:
            n_ident, n_expr = self.parse_for_assignment(rv)

        rv.set_ident(n_ident)
        if not isinstance(n_expr, Range_Expression):
            raise ICE("parfor range is not a range")
        else:
            rv.set_range(n_expr)

        self.match_eos(rv)

        rv.set_body(self.parse_delimited_input())

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)
        self.pop_context()

        return rv

    def parse_while_statement(self):
        self.match("KEYWORD", "while")
        self.push_context("loop")
        t_kw = self.ct
        with self.sc_context(True):
            n_guard = self.parse_expression()
        rv = While_Statement(t_kw, n_guard)
        self.match_eos(rv)

        rv.set_body(self.parse_delimited_input())
        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)
        self.pop_context()

        return rv

    def parse_global_statement(self):
        self.match("KEYWORD", "global")
        rv = Global_Statement(self.ct)

        while True:
            rv.add_name(self.parse_identifier(allow_void=False))
            if not self.peek("IDENTIFIER"):
                break

        self.match_eos(rv, allow_nothing=True)

        return rv

    def parse_persistent_statement(self):
        self.match("KEYWORD", "persistent")
        rv = Persistent_Statement(self.ct)

        while True:
            rv.add_name(self.parse_identifier(allow_void=False))
            if self.peek_eos():
                self.match_eos(rv)
                break

        return rv

    def parse_switch_statement(self):
        self.match("KEYWORD", "switch")
        self.push_context("switch")
        t_switch = self.ct
        n_switch_expr = self.parse_expression()
        rv = Switch_Statement(t_switch, n_switch_expr)
        self.match_eos(rv)

        while True:
            if self.peek("KEYWORD", "otherwise"):
                self.match("KEYWORD", "otherwise")
                n_action = Action(self.ct)
                self.match_eos(n_action, allow_nothing=True)
                n_action.set_body(self.parse_delimited_input())
                rv.add_action(n_action)
                break
            else:
                self.match("KEYWORD", "case")
                n_action = Action(self.ct)
                n_action.set_expression(self.parse_expression())
                self.match_eos(n_action, allow_nothing=True)
                n_action.set_body(self.parse_delimited_input())
                rv.add_action(n_action)

            if self.peek("KEYWORD", "end"):
                break

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)
        self.pop_context()

        return rv

    def parse_import_statement(self):
        # https://www.mathworks.com/help/matlab/ref/import.html
        #
        # What makes this one curious is that previously it was
        # probably a command-form function. Now it is a statement. But
        # for this reason (probably) commas are not used to separate
        # the things to import.
        #
        # In MISS_HIT for now you can only import a single name per
        # import statement (i.e. no space separated lists allowed
        # here).
        self.match("KEYWORD", "import")
        rv = Import_Statement(self.ct)

        self.match("IDENTIFIER")
        chain = [self.ct]
        while self.peek("SELECTION") or self.peek("OPERATOR", ".*"):
            if self.peek("OPERATOR", ".*"):
                self.match("OPERATOR", ".*")
                chain.append(self.ct)
                break
            else:
                self.match("SELECTION")
                self.ct.set_ast(rv)
                self.match("IDENTIFIER")
                chain.append(self.ct)

        rv.set_chain(chain)
        self.match_eos(rv)

        return rv

    def parse_try_statement(self):
        self.match("KEYWORD", "try")
        self.push_context("block")
        rv = Try_Statement(self.ct)
        self.match_eos(rv, allow_nothing=True)

        rv.set_body(self.parse_delimited_input())

        if self.peek("KEYWORD", "end"):
            # A missing catch block seems to be an undocumented
            # extension to MATLAB that Octave also supports. It should
            # be equivalent to a general catch with an empty body.
            pass

        else:
            self.match("KEYWORD", "catch")
            t_catch = self.ct
            if not self.peek_eos():
                rv.set_ident(self.parse_identifier(allow_void = False))
            self.match_eos(rv)

            rv.set_handler_body(t_catch, self.parse_delimited_input())

        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)
        self.pop_context()

        return rv

    def parse_spmd_statement(self):
        self.match("KEYWORD", "spmd")
        self.push_context("block")
        rv = SPMD_Statement(self.ct)
        self.match_eos(rv)

        rv.set_body(self.parse_delimited_input())
        self.match("KEYWORD", "end")
        self.ct.set_ast(rv)
        self.match_eos(rv)
        self.pop_context()

        return rv
